# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module contains simple statistical algorithms that are
straightforwardly implemented as a single python function (or family of
functions).

This module should generally not be used directly.  Everything in
`__all__` is imported into `astropy.stats`, and hence that package
should be used for access.
"""

import math
from collections.abc import Callable
from typing import Literal, SupportsFloat, TypeVar

import numpy as np
from numpy.typing import ArrayLike, NDArray

from astropy.utils.compat.optional_deps import HAS_BOTTLENECK, HAS_MPMATH, HAS_SCIPY

from . import _stats

__all__ = [
    "binned_binom_proportion",
    "binom_conf_interval",
    "bootstrap",
    "cdf_from_intervals",
    "fold_intervals",
    "gaussian_fwhm_to_sigma",
    "gaussian_sigma_to_fwhm",
    "histogram_intervals",
    "interval_overlap_length",
    "kuiper",
    "kuiper_false_positive_probability",
    "kuiper_two",
    "mad_std",
    "median_absolute_deviation",
    "poisson_conf_interval",
    "signal_to_noise_oir_ccd",
]

__doctest_skip__ = ["binned_binom_proportion"]
__doctest_requires__ = {
    "binom_conf_interval": ["scipy"],
    "poisson_conf_interval": ["scipy"],
}

# type for variables generated with the mpmath library
FloatLike = TypeVar("FloatLike", bound=SupportsFloat)

gaussian_sigma_to_fwhm = 2.0 * math.sqrt(2.0 * math.log(2.0))
"""
Factor with which to multiply Gaussian 1-sigma standard deviation to
convert it to full width at half maximum (FWHM).
"""

gaussian_fwhm_to_sigma = 1.0 / gaussian_sigma_to_fwhm
"""
Factor with which to multiply Gaussian full width at half maximum (FWHM)
to convert it to 1-sigma standard deviation.
"""


def binom_conf_interval(
    k: int | NDArray,
    n: int | NDArray,
    confidence_level: float = 0.68269,
    interval: Literal["wilson", "jeffreys", "flat", "wald"] = "wilson",
) -> NDArray:
    r"""Binomial proportion confidence interval given k successes,
    n trials.

    Parameters
    ----------
    k : int or numpy.ndarray
        Number of successes (0 <= ``k`` <= ``n``).
    n : int or numpy.ndarray
        Number of trials (``n`` > 0).  If both ``k`` and ``n`` are arrays,
        they must have the same shape.
    confidence_level : float, optional
        Desired probability content of interval. Default is 0.68269,
        corresponding to 1 sigma in a 1-dimensional Gaussian distribution.
        Confidence level must be in range [0, 1].
    interval : {'wilson', 'jeffreys', 'flat', 'wald'}, optional
        Formula used for confidence interval. See notes for details.  The
        ``'wilson'`` and ``'jeffreys'`` intervals generally give similar
        results, while 'flat' is somewhat different, especially for small
        values of ``n``.  ``'wilson'`` should be somewhat faster than
        ``'flat'`` or ``'jeffreys'``.  The 'wald' interval is generally not
        recommended.  It is provided for comparison purposes.  Default is
        ``'wilson'``.

    Returns
    -------
    conf_interval : ndarray
        ``conf_interval[0]`` and ``conf_interval[1]`` correspond to the lower
        and upper limits, respectively, for each element in ``k``, ``n``.

    Notes
    -----
    In situations where a probability of success is not known, it can
    be estimated from a number of trials (n) and number of
    observed successes (k). For example, this is done in Monte
    Carlo experiments designed to estimate a detection efficiency. It
    is simple to take the sample proportion of successes (k/n)
    as a reasonable best estimate of the true probability
    :math:`\epsilon`. However, deriving an accurate confidence
    interval on :math:`\epsilon` is non-trivial. There are several
    formulas for this interval (see [1]_). Four intervals are implemented
    here:

    **1. The Wilson Interval.** This interval, attributed to Wilson [2]_,
    is given by

    .. math::

        CI_{\rm Wilson} = \frac{k + \kappa^2/2}{n + \kappa^2}
        \pm \frac{\kappa n^{1/2}}{n + \kappa^2}
        ((\hat{\epsilon}(1 - \hat{\epsilon}) + \kappa^2/(4n))^{1/2}

    where :math:`\hat{\epsilon} = k / n` and :math:`\kappa` is the
    number of standard deviations corresponding to the desired
    confidence interval for a *normal* distribution (for example,
    1.0 for a confidence interval of 68.269%). For a
    confidence interval of 100(1 - :math:`\alpha`)%,

    .. math::

        \kappa = \Phi^{-1}(1-\alpha/2) = \sqrt{2}{\rm erf}^{-1}(1-\alpha).

    **2. The Jeffreys Interval.** This interval is derived by applying
    Bayes' theorem to the binomial distribution with the
    noninformative Jeffreys prior [3]_, [4]_. The noninformative Jeffreys
    prior is the Beta distribution, Beta(1/2, 1/2), which has the density
    function

    .. math::

        f(\epsilon) = \pi^{-1} \epsilon^{-1/2}(1-\epsilon)^{-1/2}.

    The justification for this prior is that it is invariant under
    reparameterizations of the binomial proportion.
    The posterior density function is also a Beta distribution: Beta(k
    + 1/2, n - k + 1/2). The interval is then chosen so that it is
    *equal-tailed*: Each tail (outside the interval) contains
    :math:`\alpha`/2 of the posterior probability, and the interval
    itself contains 1 - :math:`\alpha`. This interval must be
    calculated numerically. Additionally, when k = 0 the lower limit
    is set to 0 and when k = n the upper limit is set to 1, so that in
    these cases, there is only one tail containing :math:`\alpha`/2
    and the interval itself contains 1 - :math:`\alpha`/2 rather than
    the nominal 1 - :math:`\alpha`.

    **3. A Flat prior.** This is similar to the Jeffreys interval,
    but uses a flat (uniform) prior on the binomial proportion
    over the range 0 to 1 rather than the reparametrization-invariant
    Jeffreys prior.  The posterior density function is a Beta distribution:
    Beta(k + 1, n - k + 1).  The same comments about the nature of the
    interval (equal-tailed, etc.) also apply to this option.

    **4. The Wald Interval.** This interval is given by

    .. math::

       CI_{\rm Wald} = \hat{\epsilon} \pm
       \kappa \sqrt{\frac{\hat{\epsilon}(1-\hat{\epsilon})}{n}}

    The Wald interval gives acceptable results in some limiting
    cases. Particularly, when n is very large, and the true proportion
    :math:`\epsilon` is not "too close" to 0 or 1. However, as the
    later is not verifiable when trying to estimate :math:`\epsilon`,
    this is not very helpful. Its use is not recommended, but it is
    provided here for comparison purposes due to its prevalence in
    everyday practical statistics.

    This function requires ``scipy`` for all interval types.

    References
    ----------
    .. [1] Brown, Lawrence D.; Cai, T. Tony; DasGupta, Anirban (2001).
       "Interval Estimation for a Binomial Proportion". Statistical
       Science 16 (2): 101-133. doi:10.1214/ss/1009213286

    .. [2] Wilson, E. B. (1927). "Probable inference, the law of
       succession, and statistical inference". Journal of the American
       Statistical Association 22: 209-212.

    .. [3] Jeffreys, Harold (1946). "An Invariant Form for the Prior
       Probability in Estimation Problems". Proc. R. Soc. Lond.. A 24 186
       (1007): 453-461. doi:10.1098/rspa.1946.0056

    .. [4] Jeffreys, Harold (1998). Theory of Probability. Oxford
       University Press, 3rd edition. ISBN 978-0198503682

    Examples
    --------
    Integer inputs return an array with shape (2,):

    >>> binom_conf_interval(4, 5, interval='wilson')  # doctest: +FLOAT_CMP
    array([0.57921724, 0.92078259])

    Arrays of arbitrary dimension are supported. The Wilson and Jeffreys
    intervals give similar results, even for small k, n:

    >>> binom_conf_interval([1, 2], 5, interval='wilson')  # doctest: +FLOAT_CMP
    array([[0.07921741, 0.21597328],
           [0.42078276, 0.61736012]])

    >>> binom_conf_interval([1, 2,], 5, interval='jeffreys')  # doctest: +FLOAT_CMP
    array([[0.0842525 , 0.21789949],
           [0.42218001, 0.61753691]])

    >>> binom_conf_interval([1, 2], 5, interval='flat')  # doctest: +FLOAT_CMP
    array([[0.12139799, 0.24309021],
           [0.45401727, 0.61535699]])

    In contrast, the Wald interval gives poor results for small k, n.
    For k = 0 or k = n, the interval always has zero length.

    >>> binom_conf_interval([1, 2], 5, interval='wald')  # doctest: +FLOAT_CMP
    array([[0.02111437, 0.18091075],
           [0.37888563, 0.61908925]])

    For confidence intervals approaching 1, the Wald interval for
    0 < k < n can give intervals that extend outside [0, 1]:

    >>> binom_conf_interval([1, 2], 5, interval='wald', confidence_level=0.99)  # doctest: +FLOAT_CMP
    array([[-0.26077835, -0.16433593],
           [ 0.66077835,  0.96433593]])

    """
    if confidence_level < 0.0 or confidence_level > 1.0:
        raise ValueError("confidence_level must be between 0. and 1.")
    alpha = 1.0 - confidence_level

    k = np.asarray(k).astype(int)
    n = np.asarray(n).astype(int)

    if (n <= 0).any():
        raise ValueError("n must be positive")
    if (k < 0).any() or (k > n).any():
        raise ValueError("k must be in {0, 1, .., n}")

    if interval == "wilson" or interval == "wald":
        from scipy.special import erfinv

        kappa = np.sqrt(2.0) * min(erfinv(confidence_level), 1.0e10)  # Avoid overflows.
        k = k.astype(float)
        n = n.astype(float)
        p = k / n

        if interval == "wilson":
            midpoint = (k + kappa**2 / 2.0) / (n + kappa**2)
            halflength = (
                (kappa * np.sqrt(n))
                / (n + kappa**2)
                * np.sqrt(p * (1 - p) + kappa**2 / (4 * n))
            )
            conf_interval = np.array([midpoint - halflength, midpoint + halflength])

            # Correct intervals out of range due to floating point errors.
            conf_interval[conf_interval < 0.0] = 0.0
            conf_interval[conf_interval > 1.0] = 1.0
        else:
            midpoint = p
            halflength = kappa * np.sqrt(p * (1.0 - p) / n)
            conf_interval = np.array([midpoint - halflength, midpoint + halflength])

    elif interval == "jeffreys" or interval == "flat":
        from scipy.special import betaincinv

        if interval == "jeffreys":
            lowerbound = betaincinv(k + 0.5, n - k + 0.5, 0.5 * alpha)
            upperbound = betaincinv(k + 0.5, n - k + 0.5, 1.0 - 0.5 * alpha)
        else:
            lowerbound = betaincinv(k + 1, n - k + 1, 0.5 * alpha)
            upperbound = betaincinv(k + 1, n - k + 1, 1.0 - 0.5 * alpha)

        # Set lower or upper bound to k/n when k/n = 0 or 1
        #  We have to treat the special case of k/n being scalars,
        #  which is an ugly kludge
        if lowerbound.ndim == 0:
            if k == 0:
                lowerbound = 0.0
            elif k == n:
                upperbound = 1.0
        else:
            lowerbound[k == 0] = 0
            upperbound[k == n] = 1

        conf_interval = np.array([lowerbound, upperbound])
    else:
        raise ValueError(f"Unrecognized interval: {interval:s}")

    return conf_interval


def binned_binom_proportion(
    x: ArrayLike,
    success: ArrayLike,
    bins: int | ArrayLike = 10,
    range: tuple[float, float] | None = None,
    confidence_level: float = 0.68269,
    interval: Literal["wilson", "jeffreys", "flat", "wald"] = "wilson",
) -> tuple[NDArray, NDArray, NDArray, NDArray]:
    """Binomial proportion and confidence interval in bins of a continuous
    variable ``x``.

    Given a set of datapoint pairs where the ``x`` values are
    continuously distributed and the ``success`` values are binomial
    ("success / failure" or "true / false"), place the pairs into
    bins according to ``x`` value and calculate the binomial proportion
    (fraction of successes) and confidence interval in each bin.

    Parameters
    ----------
    x : sequence
        Values.
    success : sequence of bool
        Success (`True`) or failure (`False`) corresponding to each value
        in ``x``.  Must be same length as ``x``.
    bins : int or sequence of scalar, optional
        If bins is an int, it defines the number of equal-width bins
        in the given range (10, by default). If bins is a sequence, it
        defines the bin edges, including the rightmost edge, allowing
        for non-uniform bin widths (in this case, 'range' is ignored).
    range : (float, float), optional
        The lower and upper range of the bins. If `None` (default),
        the range is set to ``(x.min(), x.max())``. Values outside the
        range are ignored.
    confidence_level : float, optional
        Must be in range [0, 1].
        Desired probability content in the confidence
        interval ``(p - perr[0], p + perr[1])`` in each bin. Default is
        0.68269.
    interval : {'wilson', 'jeffreys', 'flat', 'wald'}, optional
        Formula used to calculate confidence interval on the
        binomial proportion in each bin. See `binom_conf_interval` for
        definition of the intervals.  The 'wilson', 'jeffreys',
        and 'flat' intervals generally give similar results.  'wilson'
        should be somewhat faster, while 'jeffreys' and 'flat' are
        marginally superior, but differ in the assumed prior.
        The 'wald' interval is generally not recommended.
        It is provided for comparison purposes. Default is 'wilson'.

    Returns
    -------
    bin_ctr : ndarray
        Central value of bins. Bins without any entries are not returned.
    bin_halfwidth : ndarray
        Half-width of each bin such that ``bin_ctr - bin_halfwidth`` and
        ``bin_ctr + bins_halfwidth`` give the left and right side of each bin,
        respectively.
    p : ndarray
        Efficiency in each bin.
    perr : ndarray
        2-d array of shape (2, len(p)) representing the upper and lower
        uncertainty on p in each bin.

    Notes
    -----
    This function requires ``scipy`` for all interval types.

    See Also
    --------
    binom_conf_interval : Function used to estimate confidence interval in
                          each bin.

    Examples
    --------
    Suppose we wish to estimate the efficiency of a survey in
    detecting astronomical sources as a function of magnitude (i.e.,
    the probability of detecting a source given its magnitude). In a
    realistic case, we might prepare a large number of sources with
    randomly selected magnitudes, inject them into simulated images,
    and then record which were detected at the end of the reduction
    pipeline. As a toy example, we generate 100 data points with
    randomly selected magnitudes between 20 and 30 and "observe" them
    with a known detection function (here, the error function, with
    50% detection probability at magnitude 25):

    >>> from scipy.special import erf
    >>> from scipy.stats.distributions import binom
    >>> def true_efficiency(x):
    ...     return 0.5 - 0.5 * erf((x - 25.) / 2.)
    >>> mag = 20. + 10. * np.random.rand(100)
    >>> detected = binom.rvs(1, true_efficiency(mag))
    >>> bins, binshw, p, perr = binned_binom_proportion(mag, detected, bins=20)
    >>> plt.errorbar(bins, p, xerr=binshw, yerr=perr, ls='none', marker='o',
    ...              label='estimate')

    .. plot::

       import numpy as np
       from scipy.special import erf
       from scipy.stats.distributions import binom
       import matplotlib.pyplot as plt
       from astropy.stats import binned_binom_proportion
       def true_efficiency(x):
           return 0.5 - 0.5 * erf((x - 25.) / 2.)
       np.random.seed(400)
       mag = 20. + 10. * np.random.rand(100)
       np.random.seed(600)
       detected = binom.rvs(1, true_efficiency(mag))
       bins, binshw, p, perr = binned_binom_proportion(mag, detected, bins=20)
       plt.errorbar(bins, p, xerr=binshw, yerr=perr, ls='none', marker='o',
                    label='estimate')
       X = np.linspace(20., 30., 1000)
       plt.plot(X, true_efficiency(X), label='true efficiency')
       plt.ylim(0., 1.)
       plt.title('Detection efficiency vs magnitude')
       plt.xlabel('Magnitude')
       plt.ylabel('Detection efficiency')
       plt.legend()
       plt.show()

    The above example uses the Wilson confidence interval to calculate
    the uncertainty ``perr`` in each bin (see the definition of various
    confidence intervals in `binom_conf_interval`). A commonly used
    alternative is the Wald interval. However, the Wald interval can
    give nonsensical uncertainties when the efficiency is near 0 or 1,
    and is therefore **not** recommended. As an illustration, the
    following example shows the same data as above but uses the Wald
    interval rather than the Wilson interval to calculate ``perr``:

    >>> bins, binshw, p, perr = binned_binom_proportion(mag, detected, bins=20,
    ...                                                 interval='wald')
    >>> plt.errorbar(bins, p, xerr=binshw, yerr=perr, ls='none', marker='o',
    ...              label='estimate')

    .. plot::

       import numpy as np
       from scipy.special import erf
       from scipy.stats.distributions import binom
       import matplotlib.pyplot as plt
       from astropy.stats import binned_binom_proportion
       def true_efficiency(x):
           return 0.5 - 0.5 * erf((x - 25.) / 2.)
       np.random.seed(400)
       mag = 20. + 10. * np.random.rand(100)
       np.random.seed(600)
       detected = binom.rvs(1, true_efficiency(mag))
       bins, binshw, p, perr = binned_binom_proportion(mag, detected, bins=20,
                                                       interval='wald')
       plt.errorbar(bins, p, xerr=binshw, yerr=perr, ls='none', marker='o',
                    label='estimate')
       X = np.linspace(20., 30., 1000)
       plt.plot(X, true_efficiency(X), label='true efficiency')
       plt.ylim(0., 1.)
       plt.title('The Wald interval can give nonsensical uncertainties')
       plt.xlabel('Magnitude')
       plt.ylabel('Detection efficiency')
       plt.legend()
       plt.show()

    """
    x = np.ravel(x)
    success = np.ravel(success).astype(bool)
    if x.shape != success.shape:
        raise ValueError("sizes of x and success must match")

    # Put values into a histogram (`n`). Put "successful" values
    # into a second histogram (`k`) with identical binning.
    n, bin_edges = np.histogram(x, bins=bins, range=range)
    k, bin_edges = np.histogram(x[success], bins=bin_edges)
    bin_ctr = (bin_edges[:-1] + bin_edges[1:]) / 2.0
    bin_halfwidth = bin_ctr - bin_edges[:-1]

    # Remove bins with zero entries.
    valid = n > 0
    bin_ctr = bin_ctr[valid]
    bin_halfwidth = bin_halfwidth[valid]
    n = n[valid]
    k = k[valid]

    p = k / n
    bounds = binom_conf_interval(
        k, n, confidence_level=confidence_level, interval=interval
    )
    perr = np.abs(bounds - p)

    return bin_ctr, bin_halfwidth, p, perr


def _check_poisson_conf_inputs(
    sigma: float,
    background: float,
    confidence_level: float,
    name: Literal[
        "root-n",
        "root-n-0",
        "pearson",
        "sherpagehrels",
        "frequentist-confidence",
        "kraft-burrows-nousek",
    ],
) -> None:
    if sigma != 1:
        raise ValueError(f"Only sigma=1 supported for interval {name}")
    if background != 0:
        raise ValueError(f"background not supported for interval {name}")
    if confidence_level is not None:
        raise ValueError(f"confidence_level not supported for interval {name}")


def poisson_conf_interval(
    n: int | NDArray,
    interval: Literal[
        "root-n",
        "root-n-0",
        "pearson",
        "sherpagehrels",
        "frequentist-confidence",
        "kraft-burrows-nousek",
    ] = "root-n",
    sigma: float = 1.0,
    background: float = 0.0,
    confidence_level: float | None = None,
) -> NDArray:
    r"""Poisson parameter confidence interval given observed counts.

    Parameters
    ----------
    n : int or numpy.ndarray
        Number of counts (0 <= ``n``).
    interval : {'root-n','root-n-0','pearson','sherpagehrels','frequentist-confidence', 'kraft-burrows-nousek'}, optional
        Formula used for confidence interval. See notes for details.
        Default is ``'root-n'``.
    sigma : float, optional
        Number of sigma for confidence interval; only supported for
        the 'frequentist-confidence' mode.
    background : float, optional
        Number of counts expected from the background; only supported for
        the 'kraft-burrows-nousek' mode. This number is assumed to be determined
        from a large region so that the uncertainty on its value is negligible.
    confidence_level : float, optional
        Confidence level between 0 and 1; only supported for the
        'kraft-burrows-nousek' mode.

    Returns
    -------
    conf_interval : ndarray
        ``conf_interval[0]`` and ``conf_interval[1]`` correspond to the lower
        and upper limits, respectively, for each element in ``n``.

    Notes
    -----
    The "right" confidence interval to use for Poisson data is a
    matter of debate. The CDF working group `recommends
    <https://web.archive.org/web/20210222093249/https://www-cdf.fnal.gov/physics/statistics/notes/pois_eb.txt>`_
    using root-n throughout, largely in the interest of
    comprehensibility, but discusses other possibilities. The ATLAS
    group also discusses several
    possibilities but concludes that no single representation is
    suitable for all cases.  The suggestion has also been `floated
    <https://ui.adsabs.harvard.edu/abs/2012EPJP..127...24A>`_ that error
    bars should be attached to theoretical predictions instead of
    observed data, which this function will not help with (but it's
    easy; then you really should use the square root of the theoretical
    prediction).

    The intervals implemented here are:

    **1. 'root-n'** This is a very widely used standard rule derived
    from the maximum-likelihood estimator for the mean of the Poisson
    process. While it produces questionable results for small n and
    outright wrong results for n=0, it is standard enough that people are
    (supposedly) used to interpreting these wonky values. The interval is

    .. math::

        CI = (n-\sqrt{n}, n+\sqrt{n})

    **2. 'root-n-0'** This is identical to the above except that where
    n is zero the interval returned is (0,1).

    **3. 'pearson'** This is an only-slightly-more-complicated rule
    based on Pearson's chi-squared rule (as `explained
    <https://web.archive.org/web/20210222093249/https://www-cdf.fnal.gov/physics/statistics/notes/pois_eb.txt>`_ by
    the CDF working group). It also has the nice feature that if your
    theory curve touches an endpoint of the interval, then your data
    point is indeed one sigma away. The interval is

    .. math::

        CI = (n+0.5-\sqrt{n+0.25}, n+0.5+\sqrt{n+0.25})

    **4. 'sherpagehrels'** This rule is used by default in the fitting
    package 'sherpa'. The `documentation
    <https://cxc.cfa.harvard.edu/sherpa/statistics/#chigehrels>`_ claims
    it is based on a numerical approximation published in `Gehrels
    (1986) <https://ui.adsabs.harvard.edu/abs/1986ApJ...303..336G>`_ but it
    does not actually appear there.  It is symmetrical, and while the
    upper limits are within about 1% of those given by
    'frequentist-confidence', the lower limits can be badly wrong. The
    interval is

    .. math::

        CI = (n-1-\sqrt{n+0.75}, n+1+\sqrt{n+0.75})

    **5. 'frequentist-confidence'** These are frequentist central
    confidence intervals:

    .. math::

        CI = (0.5 F_{\chi^2}^{-1}(\alpha;2n),
              0.5 F_{\chi^2}^{-1}(1-\alpha;2(n+1)))

    where :math:`F_{\chi^2}^{-1}` is the quantile of the chi-square
    distribution with the indicated number of degrees of freedom and
    :math:`\alpha` is the one-tailed probability of the normal
    distribution (at the point given by the parameter 'sigma'). See
    `Maxwell (2011)
    <https://ui.adsabs.harvard.edu/abs/2011arXiv1102.0822M>`_ for further
    details.

    **6. 'kraft-burrows-nousek'** This is a Bayesian approach which allows
    for the presence of a known background :math:`B` in the source signal
    :math:`N`.
    For a given confidence level :math:`CL` the confidence interval
    :math:`[S_\mathrm{min}, S_\mathrm{max}]` is given by:

    .. math::

       CL = \int^{S_\mathrm{max}}_{S_\mathrm{min}} f_{N,B}(S)dS

    where the function :math:`f_{N,B}` is:

    .. math::

       f_{N,B}(S) = C \frac{e^{-(S+B)}(S+B)^N}{N!}

    and the normalization constant :math:`C`:

    .. math::

       C = \left[ \int_0^\infty \frac{e^{-(S+B)}(S+B)^N}{N!} dS \right] ^{-1}
       = \left( \sum^N_{n=0} \frac{e^{-B}B^n}{n!}  \right)^{-1}

    See `Kraft, Burrows, and Nousek (1991)
    <https://ui.adsabs.harvard.edu/abs/1991ApJ...374..344K>`_ for further
    details.

    These formulas implement a positive, uniform prior.
    `Kraft, Burrows, and Nousek (1991)
    <https://ui.adsabs.harvard.edu/abs/1991ApJ...374..344K>`_ discuss this
    choice in more detail and show that the problem is relatively
    insensitive to the choice of prior.

    This function has an optional dependency: Either `Scipy
    <https://www.scipy.org/>`_ or `mpmath <https://mpmath.org/>`_  need
    to be available.
    This code is very intense numerically, which makes it much slower than
    the other methods, in particular for large count numbers (above 1000
    even with ``mpmath``). Fortunately, some of the other methods or a
    Gaussian approximation usually work well in this regime.

    Examples
    --------
    >>> poisson_conf_interval(np.arange(10), interval='root-n').T
    array([[  0.        ,   0.        ],
           [  0.        ,   2.        ],
           [  0.58578644,   3.41421356],
           [  1.26794919,   4.73205081],
           [  2.        ,   6.        ],
           [  2.76393202,   7.23606798],
           [  3.55051026,   8.44948974],
           [  4.35424869,   9.64575131],
           [  5.17157288,  10.82842712],
           [  6.        ,  12.        ]])

    >>> poisson_conf_interval(np.arange(10), interval='root-n-0').T
    array([[  0.        ,   1.        ],
           [  0.        ,   2.        ],
           [  0.58578644,   3.41421356],
           [  1.26794919,   4.73205081],
           [  2.        ,   6.        ],
           [  2.76393202,   7.23606798],
           [  3.55051026,   8.44948974],
           [  4.35424869,   9.64575131],
           [  5.17157288,  10.82842712],
           [  6.        ,  12.        ]])

    >>> poisson_conf_interval(np.arange(10), interval='pearson').T
    array([[  0.        ,   1.        ],
           [  0.38196601,   2.61803399],
           [  1.        ,   4.        ],
           [  1.69722436,   5.30277564],
           [  2.43844719,   6.56155281],
           [  3.20871215,   7.79128785],
           [  4.        ,   9.        ],
           [  4.8074176 ,  10.1925824 ],
           [  5.62771868,  11.37228132],
           [  6.45861873,  12.54138127]])

    >>> poisson_conf_interval(
    ...     np.arange(10), interval='frequentist-confidence').T
    array([[  0.        ,   1.84102165],
           [  0.17275378,   3.29952656],
           [  0.70818544,   4.63785962],
           [  1.36729531,   5.91818583],
           [  2.08566081,   7.16275317],
           [  2.84030886,   8.38247265],
           [  3.62006862,   9.58364155],
           [  4.41852954,  10.77028072],
           [  5.23161394,  11.94514152],
           [  6.05653896,  13.11020414]])

    >>> poisson_conf_interval(
    ...     7, interval='frequentist-confidence').T
    array([  4.41852954,  10.77028072])

    >>> poisson_conf_interval(
    ...     10, background=1.5, confidence_level=0.95,
    ...     interval='kraft-burrows-nousek').T  # doctest: +FLOAT_CMP
    array([[ 3.47894005, 16.113329533]])

    """
    if not np.isscalar(n):
        n = np.asanyarray(n)

    if interval == "root-n":
        _check_poisson_conf_inputs(sigma, background, confidence_level, interval)
        conf_interval = np.array([n - np.sqrt(n), n + np.sqrt(n)])
    elif interval == "root-n-0":
        _check_poisson_conf_inputs(sigma, background, confidence_level, interval)
        conf_interval = np.array([n - np.sqrt(n), n + np.sqrt(n)])
        if np.isscalar(n):
            if n == 0:
                conf_interval[1] = 1
        else:
            conf_interval[1, n == 0] = 1
    elif interval == "pearson":
        _check_poisson_conf_inputs(sigma, background, confidence_level, interval)
        conf_interval = np.array(
            [n + 0.5 - np.sqrt(n + 0.25), n + 0.5 + np.sqrt(n + 0.25)]
        )
    elif interval == "sherpagehrels":
        _check_poisson_conf_inputs(sigma, background, confidence_level, interval)
        conf_interval = np.array([n - 1 - np.sqrt(n + 0.75), n + 1 + np.sqrt(n + 0.75)])
    elif interval == "frequentist-confidence":
        _check_poisson_conf_inputs(1.0, background, confidence_level, interval)
        import scipy.stats

        alpha = scipy.stats.norm.sf(sigma)
        conf_interval = np.array(
            [
                0.5 * scipy.stats.chi2(2 * n).ppf(alpha),
                0.5 * scipy.stats.chi2(2 * n + 2).isf(alpha),
            ]
        )
        if np.isscalar(n):
            if n == 0:
                conf_interval[0] = 0
        else:
            conf_interval[0, n == 0] = 0
    elif interval == "kraft-burrows-nousek":
        # Deprecation warning when N is float, so we force int,
        # see https://github.com/astropy/astropy/issues/10832
        if np.isscalar(n):
            if not isinstance(n, int):
                raise TypeError("Number of counts must be integer.")
        elif not issubclass(n.dtype.type, np.integer):
            raise TypeError("Number of counts must be integer.")

        if confidence_level is None:
            raise ValueError(
                f"Set confidence_level for method {interval}. (sigma is ignored.)"
            )
        confidence_level = np.asanyarray(confidence_level)
        if np.any(confidence_level <= 0) or np.any(confidence_level >= 1):
            raise ValueError("confidence_level must be a number between 0 and 1.")
        background = np.asanyarray(background)
        if np.any(background < 0):
            raise ValueError("Background must be >= 0.")
        conf_interval = np.vectorize(_kraft_burrows_nousek, cache=True)(
            n, background, confidence_level
        )
        conf_interval = np.vstack(conf_interval)
    else:
        raise ValueError(f"Invalid method for Poisson confidence intervals: {interval}")
    return conf_interval


def median_absolute_deviation(
    data: ArrayLike,
    axis: int | tuple[int, ...] | None = None,
    func: Callable | None = None,
    ignore_nan: bool | None = False,
) -> float | NDArray:
    """
    Calculate the median absolute deviation (MAD).

    The MAD is defined as ``median(abs(a - median(a)))``.

    Parameters
    ----------
    data : array-like
        Input array or object that can be converted to an array.
    axis : None, int, or tuple of int, optional
        The axis or axes along which the MADs are computed.  The default
        (`None`) is to compute the MAD of the flattened array.
    func : callable, optional
        The function used to compute the median. Defaults to `numpy.ma.median`
        for masked arrays, otherwise to `numpy.median`.
    ignore_nan : bool, optional
        Ignore NaN values (treat them as if they are not in the array) when
        computing the median.  This will use `numpy.ma.median` if ``axis`` is
        specified, or `numpy.nanmedian` if ``axis==None`` and numpy's version
        is >1.10 because nanmedian is slightly faster in this case.

    Returns
    -------
    mad : float or `~numpy.ndarray`
        The median absolute deviation of the input array.  If ``axis``
        is `None` then a scalar will be returned, otherwise a
        `~numpy.ndarray` will be returned.

    Examples
    --------
    Generate random variates from a Gaussian distribution and return the
    median absolute deviation for that distribution::

        >>> import numpy as np
        >>> from astropy.stats import median_absolute_deviation
        >>> rand = np.random.default_rng(12345)
        >>> from numpy.random import randn
        >>> mad = median_absolute_deviation(rand.standard_normal(1000))
        >>> print(mad)    # doctest: +FLOAT_CMP
        0.6829504282771885

    See Also
    --------
    mad_std
    """
    if func is None:
        # Check if the array has a mask and if so use np.ma.median
        # See https://github.com/numpy/numpy/issues/7330 why using np.ma.median
        # for normal arrays should not be done (summary: np.ma.median always
        # returns an masked array even if the result should be scalar). (#4658)
        if isinstance(data, np.ma.MaskedArray):
            is_masked = True
            func = np.ma.median
            if ignore_nan:
                data = np.ma.masked_where(np.isnan(data), data, copy=True)
        elif ignore_nan:
            # prevent circular import
            from astropy.stats.nanfunctions import nanmedian

            is_masked = False
            func = nanmedian
        else:
            is_masked = False
            func = np.median  # drops units if result is NaN
    else:
        is_masked = None

    data = np.asanyarray(data)
    # np.nanmedian has `keepdims`, which is a good option if we're not allowing
    # user-passed functions here
    data_median = func(data, axis=axis)

    # broadcast the median array before subtraction
    if axis is not None:
        data_median = np.expand_dims(data_median, axis=axis)

    if HAS_BOTTLENECK:
        result = func(np.abs(data - data_median), axis=axis)
    else:
        result = func(np.abs(data - data_median), axis=axis, overwrite_input=True)

    if axis is None and np.ma.isMaskedArray(result):
        # return scalar version
        result = result.item()
    elif np.ma.isMaskedArray(result) and not is_masked:
        # if the input array was not a masked array, we don't want to return a
        # masked array
        result = result.filled(fill_value=np.nan)

    return result


def mad_std(
    data: ArrayLike,
    axis: int | tuple[int, ...] | None = None,
    func: Callable | None = None,
    ignore_nan: bool | None = False,
) -> float | NDArray:
    r"""
    Calculate a robust standard deviation using the `median absolute
    deviation (MAD)
    <https://en.wikipedia.org/wiki/Median_absolute_deviation>`_.

    The standard deviation estimator is given by:

    .. math::

        \sigma \approx \frac{\textrm{MAD}}{\Phi^{-1}(3/4)}
            \approx 1.4826 \ \textrm{MAD}

    where :math:`\Phi^{-1}(P)` is the normal inverse cumulative
    distribution function evaluated at probability :math:`P = 3/4`.

    Parameters
    ----------
    data : array-like
        Data array or object that can be converted to an array.
    axis : None, int, or tuple of int, optional
        The axis or axes along which the robust standard deviations are
        computed.  The default (`None`) is to compute the robust
        standard deviation of the flattened array.
    func : callable, optional
        The function used to compute the median. Defaults to `numpy.ma.median`
        for masked arrays, otherwise to `numpy.median`.
    ignore_nan : bool, optional
        Ignore NaN values (treat them as if they are not in the array) when
        computing the median.  This will use `numpy.ma.median` if ``axis`` is
        specified, or `numpy.nanmedian` if ``axis=None`` and numpy's version is
        >1.10 because nanmedian is slightly faster in this case.

    Returns
    -------
    mad_std : float or `~numpy.ndarray`
        The robust standard deviation of the input data.  If ``axis`` is
        `None` then a scalar will be returned, otherwise a
        `~numpy.ndarray` will be returned.

    Examples
    --------
    >>> import numpy as np
    >>> from astropy.stats import mad_std
    >>> rand = np.random.default_rng(12345)
    >>> madstd = mad_std(rand.normal(5, 2, (100, 100)))
    >>> print(madstd)    # doctest: +FLOAT_CMP
    1.984147963351707

    See Also
    --------
    biweight_midvariance, biweight_midcovariance, median_absolute_deviation
    """
    # NOTE: 1. / scipy.stats.norm.ppf(0.75) = 1.482602218505602
    MAD = median_absolute_deviation(data, axis=axis, func=func, ignore_nan=ignore_nan)
    return MAD * 1.482602218505602


def signal_to_noise_oir_ccd(
    t: float | NDArray,
    source_eps: float,
    sky_eps: float,
    dark_eps: float,
    rd: float,
    npix: float,
    gain: float = 1.0,
) -> float | NDArray:
    """Computes the signal to noise ratio for source being observed in the
    optical/IR using a CCD.

    Parameters
    ----------
    t : float or numpy.ndarray
        CCD integration time in seconds
    source_eps : float
        Number of electrons (photons) or DN per second in the aperture from the
        source. Note that this should already have been scaled by the filter
        transmission and the quantum efficiency of the CCD. If the input is in
        DN, then be sure to set the gain to the proper value for the CCD.
        If the input is in electrons per second, then keep the gain as its
        default of 1.0.
    sky_eps : float
        Number of electrons (photons) or DN per second per pixel from the sky
        background. Should already be scaled by filter transmission and QE.
        This must be in the same units as source_eps for the calculation to
        make sense.
    dark_eps : float
        Number of thermal electrons per second per pixel. If this is given in
        DN or ADU, then multiply by the gain to get the value in electrons.
    rd : float
        Read noise of the CCD in electrons. If this is given in
        DN or ADU, then multiply by the gain to get the value in electrons.
    npix : float
        Size of the aperture in pixels
    gain : float, optional
        Gain of the CCD. In units of electrons per DN.

    Returns
    -------
    SNR : float or numpy.ndarray
        Signal to noise ratio calculated from the inputs
    """
    signal = t * source_eps * gain
    noise = np.sqrt(
        t * (source_eps * gain + npix * (sky_eps * gain + dark_eps)) + npix * rd**2
    )
    return signal / noise


def bootstrap(
    data: NDArray,
    bootnum: int = 100,
    samples: int | None = None,
    bootfunc: Callable | None = None,
) -> NDArray:
    """Performs bootstrap resampling on numpy arrays.

    Bootstrap resampling is used to understand confidence intervals of sample
    estimates. This function returns versions of the dataset resampled with
    replacement ("case bootstrapping"). These can all be run through a function
    or statistic to produce a distribution of values which can then be used to
    find the confidence intervals.

    Parameters
    ----------
    data : ndarray
        N-D array. The bootstrap resampling will be performed on the first
        index, so the first index should access the relevant information
        to be bootstrapped.
    bootnum : int, optional
        Number of bootstrap resamples
    samples : int, optional
        Number of samples in each resample. The default `None` sets samples to
        the number of datapoints
    bootfunc : function, optional
        Function to reduce the resampled data. Each bootstrap resample will
        be put through this function and the results returned. If `None`, the
        bootstrapped data will be returned

    Returns
    -------
    boot : ndarray

        If bootfunc is None, then each row is a bootstrap resample of the data.
        If bootfunc is specified, then the columns will correspond to the
        outputs of bootfunc.

    Examples
    --------
    Obtain a twice resampled array:

    >>> from astropy.stats import bootstrap
    >>> import numpy as np
    >>> from astropy.utils import NumpyRNGContext
    >>> bootarr = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 0])
    >>> with NumpyRNGContext(1):
    ...     bootresult = bootstrap(bootarr, 2)
    ...
    >>> bootresult  # doctest: +FLOAT_CMP
    array([[6., 9., 0., 6., 1., 1., 2., 8., 7., 0.],
           [3., 5., 6., 3., 5., 3., 5., 8., 8., 0.]])
    >>> bootresult.shape
    (2, 10)

    Obtain a statistic on the array

    >>> with NumpyRNGContext(1):
    ...     bootresult = bootstrap(bootarr, 2, bootfunc=np.mean)
    ...
    >>> bootresult  # doctest: +FLOAT_CMP
    array([4. , 4.6])

    Obtain a statistic with two outputs on the array

    >>> test_statistic = lambda x: (np.sum(x), np.mean(x))
    >>> with NumpyRNGContext(1):
    ...     bootresult = bootstrap(bootarr, 3, bootfunc=test_statistic)
    >>> bootresult  # doctest: +FLOAT_CMP
    array([[40. ,  4. ],
           [46. ,  4.6],
           [35. ,  3.5]])
    >>> bootresult.shape
    (3, 2)

    Obtain a statistic with two outputs on the array, keeping only the first
    output

    >>> bootfunc = lambda x:test_statistic(x)[0]
    >>> with NumpyRNGContext(1):
    ...     bootresult = bootstrap(bootarr, 3, bootfunc=bootfunc)
    ...
    >>> bootresult  # doctest: +FLOAT_CMP
    array([40., 46., 35.])
    >>> bootresult.shape
    (3,)

    """
    if samples is None:
        samples = data.shape[0]

    # make sure the input is sane
    if samples < 1 or bootnum < 1:
        raise ValueError("neither 'samples' nor 'bootnum' can be less than 1.")

    if bootfunc is None:
        resultdims = (bootnum,) + (samples,) + data.shape[1:]
    else:
        # test number of outputs from bootfunc, avoid single outputs which are
        # array-like
        try:
            resultdims = (bootnum, len(bootfunc(data)))
        except TypeError:
            resultdims = (bootnum,)

    # create empty boot array
    boot = np.empty(resultdims)

    for i in range(bootnum):
        bootarr = np.random.randint(low=0, high=data.shape[0], size=samples)
        if bootfunc is None:
            boot[i] = data[bootarr]
        else:
            boot[i] = bootfunc(data[bootarr])

    return boot


def _scipy_kraft_burrows_nousek(N: int, B: float, CL: float) -> tuple[float, float]:
    """Upper limit on a poisson count rate.

    The implementation is based on Kraft, Burrows and Nousek
    `ApJ 374, 344 (1991) <https://ui.adsabs.harvard.edu/abs/1991ApJ...374..344K>`_.
    The XMM-Newton upper limit server uses the same formalism.

    Parameters
    ----------
    N : int or np.int32/np.int64
        Total observed count number
    B : float or np.float32/np.float64
        Background count rate (assumed to be known with negligible error
        from a large background area).
    CL : float or np.float32/np.float64
       Confidence level (number between 0 and 1)

    Returns
    -------
    S : source count limit

    Notes
    -----
    Requires :mod:`~scipy`.
    """
    from scipy.integrate import quad
    from scipy.optimize import brentq
    from scipy.special import gammaln

    B = float(B)

    def eqn8(N: int, B: float) -> float:
        n = np.arange(N + 1, dtype=np.float64)
        # version of math.exp(-B) * np.sum(np.power(B, n) / factorial(n))
        # which is safe for large N
        return 1.0 / np.sum(np.exp(n * math.log(B) - B - gammaln(n + 1)))

    # The parameters of eqn8 do not vary between calls so we can calculate the
    # result once and reuse it. The same is True for the factorial of N.
    # eqn7 is called hundred times so "caching" these values yields a
    # significant speedup (factor 10).
    eqn8_res = eqn8(N, B)

    def eqn7(S: float, N: int, B: float) -> float:
        SpB = S + B
        # We use gammaln to avoid overflow for large N.
        # It is identical to eqn8_res * (math.exp(-SpB) * SpB**N / factorial_N)
        # See https://en.wikipedia.org/wiki/Poisson_distribution#Computational_methods
        return eqn8_res * math.exp(N * math.log(SpB) - SpB - gammaln(N + 1))

    def eqn9_left(S_min: float, S_max: float, N: int, B: float) -> tuple[float, float]:
        return quad(eqn7, S_min, S_max, args=(N, B), limit=500)

    def find_s_min(S_max: float, N: int, B: float) -> float:
        """
        Kraft, Burrows and Nousek suggest to integrate from N-B in both
        directions at once, so that S_min and S_max move similarly (see
        the article for details). Here, this is implemented differently:
        Treat S_max as the optimization parameters in func and then
        calculate the matching s_min that has has eqn7(S_max) =
        eqn7(S_min) here.
        """
        y_S_max = eqn7(S_max, N, B)
        if eqn7(0, N, B) >= y_S_max:
            return 0.0
        else:
            return brentq(lambda x: eqn7(x, N, B) - y_S_max, 0, N - B)

    def func(s: float) -> float:
        s_min = find_s_min(s, N, B)
        out = eqn9_left(s_min, s, N, B)
        return out[0] - CL

    # max(100, np.sqrt(N) * 10 + N) chosen to represent ~10 sigma over input number
    # while never being smaller than 100
    S_max = brentq(func, N - B, max(100, np.sqrt(N) * 10 + N))
    S_min = find_s_min(S_max, N, B)
    return S_min, S_max


def _mpmath_kraft_burrows_nousek(N: int, B: float, CL: float) -> tuple[float, float]:
    """Upper limit on a poisson count rate.

    The implementation is based on Kraft, Burrows and Nousek in
    `ApJ 374, 344 (1991) <https://ui.adsabs.harvard.edu/abs/1991ApJ...374..344K>`_.
    The XMM-Newton upper limit server used the same formalism.

    Parameters
    ----------
    N : int or np.int32/np.int64
        Total observed count number
    B : float or np.float32/np.float64
        Background count rate (assumed to be known with negligible error
        from a large background area).
    CL : float or np.float32/np.float64
       Confidence level (number between 0 and 1)

    Returns
    -------
    S : source count limit

    Notes
    -----
    Requires the `mpmath <https://mpmath.org/>`_ library.  See
    `~astropy.stats.scipy_poisson_upper_limit` for an implementation
    that is based on scipy and evaluates faster, but runs only to about
    N = 100.
    """
    from mpmath import exp, factorial, findroot, fsum, mpf, power, quad

    # We convert these values to float. Because for some reason,
    # mpmath.mpf cannot convert from numpy.int64
    N = mpf(float(N))
    B = mpf(float(B))
    CL = mpf(float(CL))
    tol = 1e-4

    def eqn8(N: FloatLike, B: FloatLike) -> FloatLike:
        sumterms = [power(B, n) / factorial(n) for n in range(int(N) + 1)]
        return 1.0 / (exp(-B) * fsum(sumterms))

    eqn8_res = eqn8(N, B)
    factorial_N = factorial(N)

    def eqn7(S: FloatLike, N: FloatLike, B: FloatLike) -> FloatLike:
        SpB = S + B
        return eqn8_res * (exp(-SpB) * SpB**N / factorial_N)

    def eqn9_left(
        S_min: FloatLike, S_max: FloatLike, N: FloatLike, B: FloatLike
    ) -> FloatLike:
        def eqn7NB(S: FloatLike) -> FloatLike:
            return eqn7(S, N, B)

        return quad(eqn7NB, [S_min, S_max])

    def find_s_min(S_max: FloatLike, N: FloatLike, B: FloatLike) -> FloatLike:
        """
        Kraft, Burrows and Nousek suggest to integrate from N-B in both
        directions at once, so that S_min and S_max move similarly (see
        the article for details). Here, this is implemented differently:
        Treat S_max as the optimization parameters in func and then
        calculate the matching s_min that has has eqn7(S_max) =
        eqn7(S_min) here.
        """
        y_S_max = eqn7(S_max, N, B)
        # If B > N, then N-B, the "most probable" values is < 0
        # and thus s_min is certainly 0.
        # Note: For small N, s_max is also close to 0 and root finding
        # might find the wrong root, thus it is important to handle this
        # case here and return the analytical answer (s_min = 0).
        if (B >= N) or (eqn7(0, N, B) >= y_S_max):
            return 0.0
        else:

            def eqn7ysmax(x: FloatLike) -> FloatLike:
                return eqn7(x, N, B) - y_S_max

            return findroot(eqn7ysmax, [0.0, N - B], solver="ridder", tol=tol)

    def func(s: FloatLike) -> FloatLike:
        s_min = find_s_min(s, N, B)
        out = eqn9_left(s_min, s, N, B)
        return out - CL

    # Several numerical problems were found prevent the solvers from finding
    # the roots unless the starting values are very close to the final values.
    # Thus, this primitive, time-wasting, brute-force stepping here to get
    # an interval that can be fed into the ridder solver.
    s_max_guess = max(N - B, 1.0)
    while func(s_max_guess) < 0:
        s_max_guess += 1
    S_max = findroot(func, [s_max_guess - 1, s_max_guess], solver="ridder", tol=tol)
    S_min = find_s_min(S_max, N, B)
    return float(S_min), float(S_max)


def _kraft_burrows_nousek(N: int, B: float, CL: float) -> tuple[float, float]:
    """Upper limit on a poisson count rate.

    The implementation is based on Kraft, Burrows and Nousek in
    `ApJ 374, 344 (1991) <https://ui.adsabs.harvard.edu/abs/1991ApJ...374..344K>`_.
    The XMM-Newton upper limit server used the same formalism.

    Parameters
    ----------
    N : int or np.int32/np.int64
        Total observed count number
    B : float or np.float32/np.float64
        Background count rate (assumed to be known with negligible error
        from a large background area).
    CL : float or np.float32/np.float64
       Confidence level (number between 0 and 1)

    Returns
    -------
    S : source count limit

    Notes
    -----
    This functions has an optional dependency: Either :mod:`scipy` or `mpmath
    <https://mpmath.org/>`_  need to be available.
    """
    if HAS_SCIPY:
        try:
            return _scipy_kraft_burrows_nousek(N, B, CL)
        except OverflowError:
            if not HAS_MPMATH:
                raise ValueError("Need mpmath package for input numbers this large.")
    if HAS_MPMATH:
        return _mpmath_kraft_burrows_nousek(N, B, CL)

    raise ImportError("Either scipy or mpmath are required.")


def kuiper_false_positive_probability(D: float, N: float) -> float:
    """Compute the false positive probability for the Kuiper statistic.

    Uses the set of four formulas described in Paltani 2004; they report
    the resulting function never underestimates the false positive
    probability but can be a bit high in the N=40..50 range.
    (They quote a factor 1.5 at the 1e-7 level.)

    Parameters
    ----------
    D : float
        The Kuiper test score.
    N : float
        The effective sample size.

    Returns
    -------
    fpp : float
        The probability of a score this large arising from the null hypothesis.

    Notes
    -----
    Eq 7 of Paltani 2004 appears to incorrectly quote the original formula
    (Stephens 1965). This function implements the original formula, as it
    produces a result closer to Monte Carlo simulations.

    References
    ----------
    .. [1] Paltani, S., "Searching for periods in X-ray observations using
           Kuiper's test. Application to the ROSAT PSPC archive",
           Astronomy and Astrophysics, v.240, p.789-790, 2004.

    .. [2] Stephens, M. A., "The goodness-of-fit statistic VN: distribution
           and significance points", Biometrika, v.52, p.309, 1965.

    """
    if not HAS_SCIPY:
        raise ModuleNotFoundError(
            "scipy is required for kuiper_false_positive_probability"
        )
    from scipy.special import comb, factorial

    if D < 0.0 or D > 2.0:
        raise ValueError("Must have 0<=D<=2 by definition of the Kuiper test")

    if D < 2.0 / N:
        return 1.0 - factorial(N) * (D - 1.0 / N) ** (N - 1)
    elif D < 3.0 / N:
        k = -(N * D - 1.0) / 2.0
        r = np.sqrt(k**2 - (N * D - 2.0) ** 2 / 2.0)
        a, b = -k + r, -k - r
        return 1 - (
            factorial(N - 1)
            * (b ** (N - 1) * (1 - a) - a ** (N - 1) * (1 - b))
            / N ** (N - 2)
            / (b - a)
        )
    elif (D > 0.5 and N % 2 == 0) or (D > (N - 1.0) / (2.0 * N) and N % 2 == 1):
        # NOTE: the upper limit of this sum is taken from Stephens 1965
        t = np.arange(np.floor(N * (1 - D)) + 1)
        y = D + t / N
        Tt = y ** (t - 3) * (
            y**3 * N
            - y**2 * t * (3 - 2 / N)
            + y * t * (t - 1) * (3 - 2 / N) / N
            - t * (t - 1) * (t - 2) / N**2
        )
        term1 = comb(N, t)
        term2 = (1 - D - t / N) ** (N - t - 1)
        # term1 is formally finite, but is approximated by numpy as np.inf for
        # large values, so we set them to zero manually when they would be
        # multiplied by zero anyway
        term1[(term1 == np.inf) & (term2 == 0)] = 0.0
        final_term = Tt * term1 * term2
        return final_term.sum()
    else:
        z = D * np.sqrt(N)
        # When m*z>18.82 (sqrt(-log(finfo(double))/2)), exp(-2m**2z**2)
        # underflows.  Cutting off just before avoids triggering a (pointless)
        # underflow warning if `under="warn"`.
        ms = np.arange(1, 18.82 / z)
        S1 = (2 * (4 * ms**2 * z**2 - 1) * np.exp(-2 * ms**2 * z**2)).sum()
        S2 = (ms**2 * (4 * ms**2 * z**2 - 3) * np.exp(-2 * ms**2 * z**2)).sum()
        return S1 - 8 * D / 3 * S2


def kuiper(
    data: ArrayLike,
    cdf: Callable = lambda x: x,
    args: tuple | list | None = (),
) -> tuple[float, float]:
    """Compute the Kuiper statistic.

    Use the Kuiper statistic version of the Kolmogorov-Smirnov test to
    find the probability that a sample like ``data`` was drawn from the
    distribution whose CDF is given as ``cdf``.

    .. warning::
        This will not work correctly for distributions that are actually
        discrete (Poisson, for example).

    Parameters
    ----------
    data : array-like
        The data values.
    cdf : callable
        A callable to evaluate the CDF of the distribution being tested
        against. Will be called with a vector of all values at once.
        The default is a uniform distribution.
    args : list-like, optional
        Additional arguments to be supplied to cdf.

    Returns
    -------
    D : float
        The raw statistic.
    fpp : float
        The probability of a D this large arising with a sample drawn from
        the distribution whose CDF is cdf.

    Notes
    -----
    The Kuiper statistic resembles the Kolmogorov-Smirnov test in that
    it is nonparametric and invariant under reparameterizations of the data.
    The Kuiper statistic, in addition, is equally sensitive throughout
    the domain, and it is also invariant under cyclic permutations (making
    it particularly appropriate for analyzing circular data).

    Returns (D, fpp), where D is the Kuiper D number and fpp is the
    probability that a value as large as D would occur if data was
    drawn from cdf.

    .. warning::
        The fpp is calculated only approximately, and it can be
        as much as 1.5 times the true value.

    Stephens 1970 claims this is more effective than the KS at detecting
    changes in the variance of a distribution; the KS is (he claims) more
    sensitive at detecting changes in the mean.

    If cdf was obtained from data by fitting, then fpp is not correct and
    it will be necessary to do Monte Carlo simulations to interpret D.
    D should normally be independent of the shape of CDF.

    References
    ----------
    .. [1] Stephens, M. A., "Use of the Kolmogorov-Smirnov, Cramer-Von Mises
           and Related Statistics Without Extensive Tables", Journal of the
           Royal Statistical Society. Series B (Methodological), Vol. 32,
           No. 1. (1970), pp. 115-122.


    """
    data = np.sort(data)
    cdfv = cdf(data, *args)
    N = len(data)
    D = np.amax(cdfv - np.arange(N) / float(N)) + np.amax(
        (np.arange(N) + 1) / float(N) - cdfv
    )

    return D, kuiper_false_positive_probability(D, N)


def kuiper_two(data1: ArrayLike, data2: ArrayLike) -> tuple[float, float]:
    """Compute the Kuiper statistic to compare two samples.

    Parameters
    ----------
    data1 : array-like
        The first set of data values.
    data2 : array-like
        The second set of data values.

    Returns
    -------
    D : float
        The raw test statistic.
    fpp : float
        The probability of obtaining two samples this different from
        the same distribution.

    .. warning::
        The fpp is quite approximate, especially for small samples.

    """
    data1 = np.sort(data1)
    data2 = np.sort(data2)
    (n1,) = data1.shape
    (n2,) = data2.shape
    common_type = np.result_type(data1.dtype, data2.dtype)
    if not (
        np.issubdtype(common_type, np.number)
        and not np.issubdtype(common_type, np.complexfloating)
    ):
        raise ValueError("kuiper_two only accepts real inputs")
    # nans, if any, are at the end after sorting.
    if np.isnan(data1[-1]) or np.isnan(data2[-1]):
        raise ValueError("kuiper_two only accepts non-nan inputs")
    D = _stats.ks_2samp(np.asarray(data1, common_type), np.asarray(data2, common_type))
    Ne = len(data1) * len(data2) / float(len(data1) + len(data2))
    return D, kuiper_false_positive_probability(D, Ne)


def fold_intervals(
    intervals: list[tuple[float, float, float]],
) -> tuple[NDArray[float], NDArray[float]]:
    """Fold the weighted intervals to the interval (0,1).

    Convert a list of intervals (ai, bi, wi) to a list of non-overlapping
    intervals covering (0,1). Each output interval has a weight equal
    to the sum of the wis of all the intervals that include it. All intervals
    are interpreted modulo 1, and weights are accumulated counting
    multiplicity. This is appropriate, for example, if you have one or more
    blocks of observation and you want to determine how much observation
    time was spent on different parts of a system's orbit (the blocks
    should be converted to units of the orbital period first).

    Parameters
    ----------
    intervals : list of (3,) tuple
        For each tuple (ai,bi,wi); ai and bi are the limits of the interval,
        and wi is the weight to apply to the interval.

    Returns
    -------
    breaks : (N,) array of float
        The endpoints of a set of intervals covering [0,1]; breaks[0]=0 and
        breaks[-1] = 1
    weights : (N-1,) array of float
        The ith element is the sum of number of times the interval
        breaks[i],breaks[i+1] is included in each interval times the weight
        associated with that interval.

    """
    r = []
    breaks = set()
    tot = 0
    for a, b, wt in intervals:
        tot += (np.ceil(b) - np.floor(a)) * wt
        fa = a % 1
        breaks.add(fa)
        r.append((0, fa, -wt))
        fb = b % 1
        breaks.add(fb)
        r.append((fb, 1, -wt))

    breaks.add(0.0)
    breaks.add(1.0)
    breaks = sorted(breaks)
    breaks_map = {f: i for (i, f) in enumerate(breaks)}
    totals = np.zeros(len(breaks) - 1)
    totals += tot
    for a, b, wt in r:
        totals[breaks_map[a] : breaks_map[b]] += wt
    return np.array(breaks), totals


def cdf_from_intervals(breaks: NDArray[float], totals: NDArray[float]) -> Callable:
    """Construct a callable piecewise-linear CDF from a pair of arrays.

    Take a pair of arrays in the format returned by fold_intervals and
    make a callable cumulative distribution function on the interval
    (0,1).

    Parameters
    ----------
    breaks : (N,) array of float
        The boundaries of successive intervals.
    totals : (N-1,) array of float
        The weight for each interval.

    Returns
    -------
    f : callable
        A cumulative distribution function corresponding to the
        piecewise-constant probability distribution given by breaks, weights

    """
    if breaks[0] != 0 or breaks[-1] != 1:
        raise ValueError("Intervals must be restricted to [0,1]")
    if np.any(np.diff(breaks) <= 0):
        raise ValueError("Breaks must be strictly increasing")
    if np.any(totals < 0):
        raise ValueError("Total weights in each subinterval must be nonnegative")
    if np.all(totals == 0):
        raise ValueError("At least one interval must have positive exposure")
    b = breaks.copy()
    c = np.concatenate(((0,), np.cumsum(totals * np.diff(b))))
    c /= c[-1]
    return lambda x: np.interp(x, b, c, 0, 1)


def interval_overlap_length(i1: tuple[float, float], i2: tuple[float, float]) -> float:
    """Compute the length of overlap of two intervals.

    Parameters
    ----------
    i1, i2 : (float, float)
        The two intervals, (interval 1, interval 2).

    Returns
    -------
    l : float
        The length of the overlap between the two intervals.

    """
    (a, b) = i1
    (c, d) = i2
    if a < c:
        if b < c:
            return 0.0
        elif b < d:
            return b - c
        else:
            return d - c
    elif a < d:
        if b < d:
            return b - a
        else:
            return d - a
    else:
        return 0


def histogram_intervals(
    n: int, breaks: NDArray[float], totals: NDArray[float]
) -> NDArray[float]:
    """Histogram of a piecewise-constant weight function.

    This function takes a piecewise-constant weight function and
    computes the average weight in each histogram bin.

    Parameters
    ----------
    n : int
        The number of bins
    breaks : (N,) array of float
        Endpoints of the intervals in the PDF
    totals : (N-1,) array of float
        Probability densities in each bin

    Returns
    -------
    h : array of float
        The average weight for each bin

    """
    h = np.zeros(n)
    start = n * breaks[0]
    for i, total in enumerate(totals):
        end = n * breaks[i + 1]
        for j in range(math.floor(start), math.ceil(end)):
            h[j] += total * interval_overlap_length((j, j + 1), (start, end))
        start = end

    return h



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest

from astropy import units as u
from astropy.units import Quantity


def test_ignore_generic_type_annotations():
    """Test annotations that are not unit related are ignored.

    This test passes if the function works.
    """

    # one unit, one not (should be ignored)
    @u.quantity_input
    def func(x: u.m, y: str):
        return x, y

    i_q, i_str = 2 * u.m, "cool string"
    o_q, o_str = func(i_q, i_str)  # if this doesn't fail, it worked.
    assert i_q == o_q
    assert i_str == o_str


class TestQuantityUnitAnnotations:
    """Test Quantity[Unit] type annotation."""

    def test_simple_annotation(self):
        @u.quantity_input
        def func(x: Quantity[u.m], y: str):
            return x, y

        i_q, i_str = 2 * u.m, "cool string"
        o_q, o_str = func(i_q, i_str)
        assert i_q == o_q
        assert i_str == o_str

        # checks the input on the 1st arg
        with pytest.raises(u.UnitsError):
            func(1 * u.s, i_str)

        # but not the second
        o_q, o_str = func(i_q, {"not": "a string"})
        assert i_q == o_q
        assert i_str != o_str

    def test_multiple_annotation(self):
        @u.quantity_input
        def multi_func(a: Quantity[u.km]) -> Quantity[u.m]:
            return a

        i_q = 2 * u.km
        o_q = multi_func(i_q)
        assert o_q == i_q
        assert o_q.unit == u.m

    def test_optional_and_annotated(self):
        @u.quantity_input
        def opt_func(x: Quantity[u.m] | None = None) -> Quantity[u.km]:
            if x is None:
                return 1 * u.km
            return x

        i_q = 250 * u.m
        o_q = opt_func(i_q)
        assert o_q.unit == u.km
        assert o_q == i_q

        i_q = None
        o_q = opt_func(i_q)
        assert o_q == 1 * u.km

    def test_union_and_annotated(self):
        #  Union and Annotated
        @u.quantity_input
        def union_func(x: Quantity[u.m] | (Quantity[u.s] | None)):
            if x is None:
                return None
            else:
                return 2 * x

        i_q = 1 * u.m
        o_q = union_func(i_q)
        assert o_q == 2 * i_q

        i_q = 1 * u.s
        o_q = union_func(i_q)
        assert o_q == 2 * i_q

        i_q = None
        o_q = union_func(i_q)
        assert o_q is None

    def test_not_unit_or_ptype(self):
        with pytest.raises(TypeError, match="unit annotation is not"):
            Quantity["definitely not a unit"]


@pytest.mark.parametrize(
    "solarx_unit,solary_unit", [(u.arcsec, u.arcsec), ("angle", "angle")]
)
def test_args3(solarx_unit, solary_unit):
    @u.quantity_input
    def myfunc_args(solarx: solarx_unit, solary: solary_unit):
        return solarx, solary

    solarx, solary = myfunc_args(1 * u.arcsec, 1 * u.arcsec)

    assert isinstance(solarx, Quantity)
    assert isinstance(solary, Quantity)

    assert solarx.unit == u.arcsec
    assert solary.unit == u.arcsec


@pytest.mark.parametrize(
    "solarx_unit,solary_unit", [(u.arcsec, u.arcsec), ("angle", "angle")]
)
def test_args_noconvert3(solarx_unit, solary_unit):
    @u.quantity_input()
    def myfunc_args(solarx: solarx_unit, solary: solary_unit):
        return solarx, solary

    solarx, solary = myfunc_args(1 * u.deg, 1 * u.arcmin)

    assert isinstance(solarx, Quantity)
    assert isinstance(solary, Quantity)

    assert solarx.unit == u.deg
    assert solary.unit == u.arcmin


@pytest.mark.parametrize("solarx_unit", [u.arcsec, "angle"])
def test_args_nonquantity3(solarx_unit):
    @u.quantity_input
    def myfunc_args(solarx: solarx_unit, solary):
        return solarx, solary

    solarx, solary = myfunc_args(1 * u.arcsec, 100)

    assert isinstance(solarx, Quantity)
    assert isinstance(solary, int)

    assert solarx.unit == u.arcsec


@pytest.mark.parametrize(
    "solarx_unit,solary_unit", [(u.arcsec, u.eV), ("angle", "energy")]
)
def test_arg_equivalencies3(solarx_unit, solary_unit):
    @u.quantity_input(equivalencies=u.mass_energy())
    def myfunc_args(solarx: solarx_unit, solary: solary_unit):
        return solarx, solary + (10 * u.J)  # Add an energy to check equiv is working

    solarx, solary = myfunc_args(1 * u.arcsec, 100 * u.gram)

    assert isinstance(solarx, Quantity)
    assert isinstance(solary, Quantity)

    assert solarx.unit == u.arcsec
    assert solary.unit == u.gram


@pytest.mark.parametrize(
    "solarx_unit,solary_unit", [(u.arcsec, u.deg), ("angle", "angle")]
)
def test_wrong_unit3(solarx_unit, solary_unit):
    @u.quantity_input
    def myfunc_args(solarx: solarx_unit, solary: solary_unit):
        return solarx, solary

    with pytest.raises(
        u.UnitsError,
        match=(
            "Argument 'solary' to function 'myfunc_args' must be in units "
            f"convertible to '{str(solary_unit)}'."
        ),
    ):
        solarx, solary = myfunc_args(1 * u.arcsec, 100 * u.km)


@pytest.mark.parametrize(
    "solarx_unit,solary_unit", [(u.arcsec, u.deg), ("angle", "angle")]
)
def test_not_quantity3(solarx_unit, solary_unit):
    @u.quantity_input
    def myfunc_args(solarx: solarx_unit, solary: solary_unit):
        return solarx, solary

    with pytest.raises(
        TypeError,
        match=(
            "Argument 'solary' to function 'myfunc_args' has no 'unit' "
            "attribute. You should pass in an astropy Quantity instead."
        ),
    ):
        solarx, solary = myfunc_args(1 * u.arcsec, 100)


def test_decorator_override():
    @u.quantity_input(solarx=u.arcsec)
    def myfunc_args(solarx: u.km, solary: u.arcsec):
        return solarx, solary

    solarx, solary = myfunc_args(1 * u.arcsec, 1 * u.arcsec)

    assert isinstance(solarx, Quantity)
    assert isinstance(solary, Quantity)

    assert solarx.unit == u.arcsec
    assert solary.unit == u.arcsec


@pytest.mark.parametrize(
    "solarx_unit,solary_unit", [(u.arcsec, u.deg), ("angle", "angle")]
)
def test_kwargs3(solarx_unit, solary_unit):
    @u.quantity_input
    def myfunc_args(solarx: solarx_unit, solary, myk: solary_unit = 1 * u.arcsec):
        return solarx, solary, myk

    solarx, solary, myk = myfunc_args(1 * u.arcsec, 100, myk=100 * u.deg)

    assert isinstance(solarx, Quantity)
    assert isinstance(solary, int)
    assert isinstance(myk, Quantity)

    assert myk.unit == u.deg


@pytest.mark.parametrize(
    "solarx_unit,solary_unit", [(u.arcsec, u.deg), ("angle", "angle")]
)
def test_unused_kwargs3(solarx_unit, solary_unit):
    @u.quantity_input
    def myfunc_args(
        solarx: solarx_unit, solary, myk: solary_unit = 1 * u.arcsec, myk2=1000
    ):
        return solarx, solary, myk, myk2

    solarx, solary, myk, myk2 = myfunc_args(1 * u.arcsec, 100, myk=100 * u.deg, myk2=10)

    assert isinstance(solarx, Quantity)
    assert isinstance(solary, int)
    assert isinstance(myk, Quantity)
    assert isinstance(myk2, int)

    assert myk.unit == u.deg
    assert myk2 == 10


@pytest.mark.parametrize("solarx_unit,energy", [(u.arcsec, u.eV), ("angle", "energy")])
def test_kwarg_equivalencies3(solarx_unit, energy):
    @u.quantity_input(equivalencies=u.mass_energy())
    def myfunc_args(solarx: solarx_unit, energy: energy = 10 * u.eV):
        return solarx, energy + (10 * u.J)  # Add an energy to check equiv is working

    solarx, energy = myfunc_args(1 * u.arcsec, 100 * u.gram)

    assert isinstance(solarx, Quantity)
    assert isinstance(energy, Quantity)

    assert solarx.unit == u.arcsec
    assert energy.unit == u.gram


@pytest.mark.parametrize(
    "solarx_unit,solary_unit", [(u.arcsec, u.deg), ("angle", "angle")]
)
def test_kwarg_wrong_unit3(solarx_unit, solary_unit):
    @u.quantity_input
    def myfunc_args(solarx: solarx_unit, solary: solary_unit = 10 * u.deg):
        return solarx, solary

    with pytest.raises(
        u.UnitsError,
        match=(
            "Argument 'solary' to function 'myfunc_args' must be in "
            f"units convertible to '{str(solary_unit)}'."
        ),
    ):
        solarx, solary = myfunc_args(1 * u.arcsec, solary=100 * u.km)


@pytest.mark.parametrize(
    "solarx_unit,solary_unit", [(u.arcsec, u.deg), ("angle", "angle")]
)
def test_kwarg_not_quantity3(solarx_unit, solary_unit):
    @u.quantity_input
    def myfunc_args(solarx: solarx_unit, solary: solary_unit = 10 * u.deg):
        return solarx, solary

    with pytest.raises(
        TypeError,
        match=(
            "Argument 'solary' to function 'myfunc_args' has no 'unit' attribute. "
            "You should pass in an astropy Quantity instead."
        ),
    ):
        solarx, solary = myfunc_args(1 * u.arcsec, solary=100)


@pytest.mark.parametrize(
    "solarx_unit,solary_unit", [(u.arcsec, u.deg), ("angle", "angle")]
)
def test_kwarg_default3(solarx_unit, solary_unit):
    @u.quantity_input
    def myfunc_args(solarx: solarx_unit, solary: solary_unit = 10 * u.deg):
        return solarx, solary

    solarx, solary = myfunc_args(1 * u.arcsec)


def test_return_annotation():
    @u.quantity_input
    def myfunc_args(solarx: u.arcsec) -> u.deg:
        return solarx

    solarx = myfunc_args(1 * u.arcsec)
    assert solarx.unit is u.deg


def test_return_annotation_none():
    @u.quantity_input
    def myfunc_args(solarx: u.arcsec) -> None:
        pass

    solarx = myfunc_args(1 * u.arcsec)
    assert solarx is None


def test_return_annotation_notUnit():
    @u.quantity_input
    def myfunc_args(solarx: u.arcsec) -> int:
        return 0

    solarx = myfunc_args(1 * u.arcsec)
    assert solarx == 0


def test_enum_annotation():
    # Regression test for gh-9932
    from enum import Enum, auto

    class BasicEnum(Enum):
        AnOption = auto()

    @u.quantity_input
    def myfunc_args(a: BasicEnum, b: u.arcsec) -> None:
        pass

    myfunc_args(BasicEnum.AnOption, 1 * u.arcsec)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Tests for miscellaneous functionality in the `funcs` module
"""

import numpy as np
import pytest
from numpy import testing as npt

from astropy import units as u
from astropy.coordinates import FK5, ICRS, SkyCoord
from astropy.coordinates import representation as r
from astropy.coordinates.funcs import (
    concatenate,
    concatenate_representations,
    get_constellation,
    get_sun,
)
from astropy.time import Time
from astropy.utils.exceptions import AstropyPendingDeprecationWarning

CARTESIAN_POS = r.CartesianRepresentation([1, 2, 3] * u.kpc)
CARTESIAN_VEL = r.CartesianDifferential([8, 9, 10] * u.km / u.s)
CARTESIAN_POS_AND_VEL = CARTESIAN_POS.with_differentials(CARTESIAN_VEL)

RADIAL_VEL = r.RadialDifferential(1 * u.km / u.s)
SPHERICAL_COS_LAT_VEL = r.SphericalCosLatDifferential(
    1 * u.mas / u.yr, 2 * u.mas / u.yr, 3 * u.km / u.s
)
SPHERICAL_POS = r.SphericalRepresentation(
    lon=1 * u.deg, lat=2.0 * u.deg, distance=10 * u.pc
)
UNIT_SPHERICAL_POS = r.UnitSphericalRepresentation(lon=1 * u.deg, lat=2.0 * u.deg)
CARTESIAN_POS_2D_ARR = r.CartesianRepresentation(np.ones((3, 100)) * u.kpc)
CARTESIAN_POS_3D_ARR = r.CartesianRepresentation(np.ones((3, 16, 8)) * u.kpc)
UNIT_SPHERICAL_COS_LAT_VEL = r.UnitSphericalCosLatDifferential(
    1 * u.mas / u.yr, 2 * u.mas / u.yr
)
CARTESIAN_VEL_2D_ARR = r.CartesianDifferential(*np.ones((3, 100)) * u.km / u.s)
CARTESIAN_VEL_3D_ARR = r.CartesianDifferential(*np.ones((3, 16, 8)) * u.km / u.s)


def test_sun():
    """
    Test that `get_sun` works and it behaves roughly as it should (in GCRS)
    """

    northern_summer_solstice = Time("2010-6-21")
    northern_winter_solstice = Time("2010-12-21")
    equinox_1 = Time("2010-3-21")
    equinox_2 = Time("2010-9-21")

    gcrs1 = get_sun(equinox_1)
    assert np.abs(gcrs1.dec.deg) < 1

    gcrs2 = get_sun(
        Time([northern_summer_solstice, equinox_2, northern_winter_solstice])
    )
    assert np.all(np.abs(gcrs2.dec - [23.5, 0, -23.5] * u.deg) < 1 * u.deg)


def test_constellations(recwarn):
    inuma = ICRS(9 * u.hour, 65 * u.deg)

    n_prewarn = len(recwarn)
    res = get_constellation(inuma)
    res_short = get_constellation(inuma, short_name=True)
    assert len(recwarn) == n_prewarn  # neither version should not make warnings

    assert res == "Ursa Major"
    assert res_short == "UMa"
    assert isinstance(res, str) or getattr(res, "shape", None) == ()

    # these are taken from the ReadMe for Roman 1987
    ras = [9, 23.5, 5.12, 9.4555, 12.8888, 15.6687, 19, 6.2222]
    decs = [65, -20, 9.12, -19.9, 22, -12.1234, -40, -81.1234]
    shortnames = ["UMa", "Aqr", "Ori", "Hya", "Com", "Lib", "CrA", "Men"]

    testcoos = FK5(ras * u.hour, decs * u.deg, equinox="B1950")
    npt.assert_equal(get_constellation(testcoos, short_name=True), shortnames)

    # test on a SkyCoord, *and* test Boötes, which is special in that it has a
    # non-ASCII character
    boores = get_constellation(SkyCoord(15 * u.hour, 30 * u.deg, frame="icrs"))
    assert boores == "Boötes"
    assert isinstance(boores, str) or getattr(boores, "shape", None) == ()


@pytest.mark.xfail
def test_constellation_edge_cases():
    # Test edge cases close to borders, using B1875.0 coordinates
    # Look for HMS / DMS roundoff-to-decimal issues from Roman (1987) data,
    # and misuse of PrecessedGeocentric, as documented in
    # https://github.com/astropy/astropy/issues/9855
    # Define eight test points.
    # The first four cross the boundary at 06h14m30 == 6.2416666666666... hours
    # with Monoceros on the west side of Orion at Dec +3.0.
    ras = [6.24100, 6.24160, 6.24166, 6.24171]
    # aka ['6h14m27.6s' '6h14m29.76s' '6h14m29.976s' '6h14m30.156s']

    decs = [3.0, 3.0, 3.0, 3.0]

    # Correct constellations for given RA/Dec coordinates
    shortnames = ["Ori", "Ori", "Ori", "Mon"]

    # The second four sample northward along RA 22 hours, crossing the boundary
    # at 86° 10' == 86.1666... degrees between Cepheus and Ursa Minor
    decs += [86.16, 86.1666, 86.16668, 86.1668]
    ras += [22.0, 22.0, 22.0, 22.0]
    shortnames += ["Cep", "Cep", "Umi", "Umi"]

    testcoos = FK5(ras * u.hour, decs * u.deg, equinox="B1875")
    npt.assert_equal(
        get_constellation(testcoos, short_name=True),
        shortnames,
        "get_constellation() error: misusing Roman approximations, vs IAU boundaries"
        " from Delporte?",
    )

    # TODO: When that's fixed, add other tests with coords that are in different constellations
    # depending on equinox


def test_concatenate():
    # Just positions
    fk5 = FK5(1 * u.deg, 2 * u.deg)
    sc = SkyCoord(3 * u.deg, 4 * u.deg, frame="fk5")

    with pytest.warns(AstropyPendingDeprecationWarning, match=r"np\.concatenate"):
        res = concatenate([fk5, sc])
    np.testing.assert_allclose(res.ra, [1, 3] * u.deg)
    np.testing.assert_allclose(res.dec, [2, 4] * u.deg)

    with (
        pytest.raises(TypeError),
        pytest.warns(AstropyPendingDeprecationWarning, match=r"np\.concatenate"),
    ):
        concatenate(fk5)

    with (
        pytest.raises(TypeError),
        pytest.warns(AstropyPendingDeprecationWarning, match=r"np\.concatenate"),
    ):
        concatenate(1 * u.deg)

    # positions and velocities
    fr = ICRS(
        ra=10 * u.deg,
        dec=11.0 * u.deg,
        pm_ra_cosdec=12 * u.mas / u.yr,
        pm_dec=13 * u.mas / u.yr,
    )
    sc = SkyCoord(
        ra=20 * u.deg,
        dec=21.0 * u.deg,
        pm_ra_cosdec=22 * u.mas / u.yr,
        pm_dec=23 * u.mas / u.yr,
    )

    with pytest.warns(AstropyPendingDeprecationWarning, match=r"np\.concatenate"):
        res = concatenate([fr, sc])

    with (
        pytest.raises(ValueError),
        pytest.warns(AstropyPendingDeprecationWarning, match=r"np\.concatenate"),
    ):
        concatenate([fr, fk5])

    fr2 = ICRS(ra=10 * u.deg, dec=11.0 * u.deg)
    with (
        pytest.raises(ValueError),
        pytest.warns(AstropyPendingDeprecationWarning, match=r"np\.concatenate"),
    ):
        concatenate([fr, fr2])


@pytest.mark.parametrize(
    "rep",
    (
        CARTESIAN_POS,
        SPHERICAL_POS,
        UNIT_SPHERICAL_POS,
        CARTESIAN_POS_2D_ARR,
        CARTESIAN_POS_3D_ARR,
        CARTESIAN_POS_AND_VEL,
        SPHERICAL_POS.with_differentials(SPHERICAL_COS_LAT_VEL),
        UNIT_SPHERICAL_POS.with_differentials(SPHERICAL_COS_LAT_VEL),
        UNIT_SPHERICAL_POS.with_differentials(UNIT_SPHERICAL_COS_LAT_VEL),
        UNIT_SPHERICAL_POS.with_differentials({"s": RADIAL_VEL}),
        CARTESIAN_POS_2D_ARR.with_differentials(CARTESIAN_VEL_2D_ARR),
        CARTESIAN_POS_3D_ARR.with_differentials(CARTESIAN_VEL_3D_ARR),
    ),
)
@pytest.mark.parametrize("n", (2, 4))
def test_concatenate_representations(rep, n):
    # Test that combining with itself succeeds
    expected_shape = (n * rep.shape[0],) + rep.shape[1:] if rep.shape else (n,)
    with pytest.warns(AstropyPendingDeprecationWarning, match=r"np\.concatenate"):
        tmp = concatenate_representations(n * (rep,))
    assert tmp.shape == expected_shape

    if "s" in rep.differentials:
        assert tmp.differentials["s"].shape == expected_shape


def test_concatenate_representations_invalid_input():
    # Test that combining pairs fails
    with (
        pytest.raises(TypeError),
        pytest.warns(AstropyPendingDeprecationWarning, match=r"np\.concatenate"),
    ):
        concatenate_representations((CARTESIAN_POS, SPHERICAL_POS))

    with (
        pytest.raises(ValueError),
        pytest.warns(AstropyPendingDeprecationWarning, match=r"np\.concatenate"),
    ):
        concatenate_representations((CARTESIAN_POS, CARTESIAN_POS_AND_VEL))

    # Check that passing in a single object fails
    with (
        pytest.raises(TypeError),
        pytest.warns(AstropyPendingDeprecationWarning, match=r"np\.concatenate"),
    ):
        concatenate_representations(CARTESIAN_POS)


def test_concatenate_representations_different_units():
    with pytest.warns(AstropyPendingDeprecationWarning, match=r"np\.concatenate"):
        concat = concatenate_representations(
            [r.CartesianRepresentation([1, 2, 3] * unit) for unit in (u.pc, u.kpc)]
        )
    assert np.array_equal(concat.xyz, [[1, 1000], [2, 2000], [3, 3000]] * u.pc)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This module contains functions to determine where configuration and
data/cache files used by Astropy should be placed.
"""

import os
import shutil
import sys
from collections.abc import Callable
from functools import partial, wraps
from inspect import cleandoc
from pathlib import Path
from types import TracebackType
from typing import Literal, ParamSpec
from warnings import warn

from astropy.utils.exceptions import AstropyUserWarning

__all__ = [
    "get_cache_dir",
    "get_cache_dir_path",
    "get_config_dir",
    "get_config_dir_path",
    "set_temp_cache",
    "set_temp_config",
]


P = ParamSpec("P")


def get_config_dir_path(rootname: str = "astropy") -> Path:
    """
    Determines the package configuration directory name and creates the
    directory if it doesn't exist.

    This directory is typically ``$HOME/.astropy/config``, but if the
    XDG_CONFIG_HOME environment variable is set and the
    ``$XDG_CONFIG_HOME/astropy`` directory exists, it will be that directory.
    If neither exists, the former will be created and symlinked to the latter.

    Parameters
    ----------
    rootname : str
        Name of the root configuration directory. For example, if ``rootname =
        'pkgname'``, the configuration directory would be ``<home>/.pkgname/``
        rather than ``<home>/.astropy`` (depending on platform).

    Returns
    -------
    configdir : Path
        The absolute path to the configuration directory.

    """
    return set_temp_config._get_dir_path(rootname)


def get_config_dir(rootname: str = "astropy") -> str:
    return str(get_config_dir_path(rootname))


if get_config_dir_path.__doc__ is not None:
    # guard against PYTHONOPTIMIZE mode
    get_config_dir.__doc__ = cleandoc(
        get_config_dir_path.__doc__
        + """
    See Also
    --------
    get_config_dir_path : same as this function, except that the return value is a pathlib.Path

    """
    )


def get_cache_dir_path(rootname: str = "astropy") -> Path:
    """
    Determines the Astropy cache directory name and creates the directory if it
    doesn't exist.

    This directory is typically ``$HOME/.astropy/cache``, but if the
    XDG_CACHE_HOME environment variable is set and the
    ``$XDG_CACHE_HOME/astropy`` directory exists, it will be that directory.
    If neither exists, the former will be created and symlinked to the latter.

    Parameters
    ----------
    rootname : str
        Name of the root cache directory. For example, if
        ``rootname = 'pkgname'``, the cache directory will be
        ``<cache>/.pkgname/``.

    Returns
    -------
    cachedir : Path
        The absolute path to the cache directory.

    """
    return set_temp_cache._get_dir_path(rootname)


def get_cache_dir(rootname: str = "astropy") -> str:
    return str(get_cache_dir_path(rootname))


if get_cache_dir_path.__doc__ is not None:
    # guard against PYTHONOPTIMIZE mode
    get_cache_dir.__doc__ = cleandoc(
        get_cache_dir_path.__doc__
        + """
    See Also
    --------
    get_cache_dir_path : same as this function, except that the return value is a pathlib.Path

    """
    )


class _SetTempPath:
    _temp_path: Path | None = None

    # This base class serves as a deduplication layer for its only two intended
    # children (set_temp_cache and set_temp_config)
    _directory_type: Literal["cache", "config"]
    _directory_env_var: Literal["XDG_CACHE_HOME", "XDG_CONFIG_HOME"]

    def __init__(
        self, path: os.PathLike[str] | str | None = None, delete: bool = False
    ) -> None:
        if path is not None:
            path = Path(path).resolve()

        self._path = path
        self._delete = delete
        self._prev_path = self.__class__._temp_path

    def __enter__(self) -> str:
        self.__class__._temp_path = self._path
        try:
            return str(self.__class__._get_dir_path(rootname="astropy"))
        except Exception:
            self.__class__._temp_path = self._prev_path
            raise

    def __exit__(
        self,
        type: type[BaseException] | None,
        value: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        self.__class__._temp_path = self._prev_path

        if self._delete and self._path is not None:
            shutil.rmtree(self._path)

    def __call__(self, func: Callable[P, object]) -> Callable[P, None]:
        """Implements use as a decorator."""

        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> None:
            with self:
                func(*args, **kwargs)

        return wrapper

    @staticmethod
    def _warn_env_var_is_ignored(
        reason: str, *, env_var_name: str, stacklevel: int = 4
    ) -> None:
        env_var_value = os.getenv(env_var_name)
        warning_msg_template = (
            f"{env_var_name} is set to '{env_var_value}', "
            "but {reason}. This environment variable will be ignored."
        )
        warn(
            warning_msg_template.format(reason=reason),
            category=AstropyUserWarning,
            stacklevel=stacklevel,
        )

    @classmethod
    def _get_dir_path(cls, rootname: str) -> Path:
        if (xch := cls._temp_path) is not None:
            path = xch / rootname
            if not path.is_file():
                path.mkdir(exist_ok=True)
            return path.resolve()

        if (env_dir_str := os.getenv(cls._directory_env_var)) is None:
            return cls._find_or_create_root_dir(linkto=None, pkgname=rootname)

        warn = partial(
            cls._warn_env_var_is_ignored, env_var_name=cls._directory_env_var
        )
        if not (env_dir_path := Path(env_dir_str)).is_dir():
            warn("no such directory was found")
            return cls._find_or_create_root_dir(linkto=None, pkgname=rootname)

        path = env_dir_path / rootname
        if path.is_symlink():
            return cls._find_or_create_root_dir(linkto=None, pkgname=rootname)
        elif path.is_dir():
            return path.resolve()

        linkto: Path | None
        if path.is_file():
            warn(
                f"this directory already contains a file under '{path.name}', "
                "where a directory was expected"
            )
            # renounce linking early to avoid redundant warnings
            linkto = None
        else:
            # symlink will be set to this if the directory is created
            linkto = path

        return cls._find_or_create_root_dir(linkto=linkto, pkgname=rootname)

    @classmethod
    def _find_or_create_root_dir(
        cls,
        linkto: Path | None,
        pkgname: str = "astropy",
    ) -> Path:
        innerdir = Path.home() / f".{pkgname}"
        maindir = innerdir / cls._directory_type

        if maindir.is_file():
            raise OSError(
                f"Intended {pkgname} {cls._directory_type} directory {maindir} is actually a file."
            )

        warn = partial(
            cls._warn_env_var_is_ignored, env_var_name=cls._directory_env_var
        )

        # At this point, linkto is either None or a path that doesn't exist yet
        # Other linkto conditions already filtered in _get_dir_path() above.
        if linkto is not None:
            if sys.platform.startswith("win"):
                warn("is not supported on Windows")
                linkto = None
            elif maindir.is_dir():
                warn(
                    f"the default location, {maindir}, already exists, and takes precedence"
                )
                linkto = None  # redundant, but shouldn't make a difference

        if not maindir.is_dir():
            # first create .astropy dir if needed
            if innerdir.is_file():
                raise OSError(
                    f"Intended {pkgname} {cls._directory_type} directory {maindir} is actually a file."
                )
            maindir.mkdir(parents=True, exist_ok=True)
            if linkto is not None:
                linkto.symlink_to(maindir)

        return maindir.resolve()


class set_temp_config(_SetTempPath):
    """
    Context manager to set a temporary path for the Astropy config, primarily
    for use with testing.

    If the path set by this context manager does not already exist it will be
    created, if possible.

    This may also be used as a decorator on a function to set the config path
    just within that function.

    Parameters
    ----------
    path : str, optional
        The directory (which must exist) in which to find the Astropy config
        files, or create them if they do not already exist.  If None, this
        restores the config path to the user's default config path as returned
        by `get_config_dir` as though this context manager were not in effect
        (this is useful for testing).  In this case the ``delete`` argument is
        always ignored.

    delete : bool, optional
        If True, cleans up the temporary directory after exiting the temp
        context (default: False).
    """

    _directory_type = "config"
    _directory_env_var = "XDG_CONFIG_HOME"

    def __enter__(self) -> str:
        # Special case for the config case, where we need to reset all the
        # cached config objects.  We do keep the cache, since some of it
        # may have been set programmatically rather than be stored in the
        # config file (e.g., iers.conf.auto_download=False for our tests).
        from .configuration import _cfgobjs

        self._cfgobjs_copy = _cfgobjs.copy()
        _cfgobjs.clear()
        return super().__enter__()

    def __exit__(
        self,
        type: type[BaseException] | None,
        value: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        from .configuration import _cfgobjs

        _cfgobjs.clear()
        _cfgobjs.update(self._cfgobjs_copy)
        del self._cfgobjs_copy
        super().__exit__(type, value, tb)


class set_temp_cache(_SetTempPath):
    """
    Context manager to set a temporary path for the Astropy download cache,
    primarily for use with testing (though there may be other applications
    for setting a different cache directory, for example to switch to a cache
    dedicated to large files).

    If the path set by this context manager does not already exist it will be
    created, if possible.

    This may also be used as a decorator on a function to set the cache path
    just within that function.

    Parameters
    ----------
    path : str
        The directory (which must exist) in which to find the Astropy cache
        files, or create them if they do not already exist.  If None, this
        restores the cache path to the user's default cache path as returned
        by `get_cache_dir` as though this context manager were not in effect
        (this is useful for testing).  In this case the ``delete`` argument is
        always ignored.

    delete : bool, optional
        If True, cleans up the temporary directory after exiting the temp
        context (default: False).
    """

    _directory_type = "cache"
    _directory_env_var = "XDG_CACHE_HOME"



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Checks for optional dependencies using lazy import from
`PEP 562 <https://www.python.org/dev/peps/pep-0562/>`_.
"""

from importlib.util import find_spec

# First, the top-level packages:
# TODO: This list is a duplicate of the dependencies in pyproject.toml "all", but
# some of the package names are different from the pip-install name (e.g.,
# beautifulsoup4 -> bs4).
# Some optional parts of the standard library also find a place here,
# but don't appear in pyproject.toml
_optional_deps = [
    "asdf_astropy",
    "bleach",
    "bottleneck",
    "bs4",
    "bz2",  # stdlib
    "certifi",
    "dask",
    "duckdb",
    "fsspec",
    "h5py",
    "html5lib",
    "ipykernel",
    "IPython",
    "ipywidgets",
    "ipydatagrid",
    "jplephem",
    "lxml",
    "matplotlib",
    "mpmath",
    "narwhals",
    "pandas",
    "PIL",
    "polars",
    "pytz",
    "s3fs",
    "scipy",
    "skyfield",
    "sortedcontainers",
    "uncompresspy",
    "lzma",  # stdlib
    "pyarrow",
    "pytest_mpl",
    "array_api_strict",
]
_deps = {k.upper(): k for k in _optional_deps}

# Any subpackages that have different import behavior:
_deps["PLT"] = "matplotlib"

__all__ = [f"HAS_{pkg}" for pkg in _deps]


def __getattr__(name):
    if name in __all__:
        return find_spec(_deps[name.removeprefix("HAS_")]) is not None

    raise AttributeError(f"Module {__name__!r} has no attribute {name!r}.")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""I/O subpackage for cosmology.

This subpackage contains classes and functions for reading, writing
and converting cosmology objects to and from various formats.

User access to the I/O functionality is provided through the methods
on the `~astropy.cosmology.Cosmology` class (and its subclasses) and its instances:

- |Cosmology.read| for reading from a file,
- |Cosmology.write| for writing to a file,
- |Cosmology.from_format| to construct a Cosmology from an object
- |Cosmology.to_format| to convert a Cosmology to an object
"""

__all__ = [
    "CosmologyFromFormat",
    "CosmologyRead",
    "CosmologyToFormat",
    "CosmologyWrite",
    "convert_registry",
    "readwrite_registry",
]

# Importing the I/O subpackage registers the I/O methods.
from ._src.io.connect import (
    CosmologyFromFormat,
    CosmologyRead,
    CosmologyToFormat,
    CosmologyWrite,
    convert_registry,
    readwrite_registry,
)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module contains functions for matching coordinate catalogs.
"""

from typing import NamedTuple

import numpy as np
from numpy.typing import NDArray

from astropy.units import Quantity

from . import Angle
from .representation import UnitSphericalRepresentation
from .sky_coordinate import SkyCoord

__all__ = [
    "CoordinateMatchResult",
    "CoordinateSearchResult",
    "match_coordinates_3d",
    "match_coordinates_sky",
    "search_around_3d",
    "search_around_sky",
]


class CoordinateMatchResult(NamedTuple):
    """Results of matching a set of sources to a catalog.

    See Also
    --------
    astropy.coordinates.match_coordinates_3d
    astropy.coordinates.match_coordinates_sky
    SkyCoord.match_to_catalog_3d
    SkyCoord.match_to_catalog_sky
    """

    indices_to_catalog: NDArray[np.int32] | NDArray[np.int64]
    """For each source the index of the match in the catalog."""
    angular_separation: Angle
    """The angular separation between each source and its match
    in the catalog."""
    physical_separation: Quantity
    """The physical separation between each source and its match in the
    catalog. If the sources or the catalog lack distances then the
    physical separations are computed assuming all coordinates are
    points on the unit sphere."""


def match_coordinates_3d(
    matchcoord, catalogcoord, nthneighbor=1, storekdtree="kdtree_3d"
):
    """
    Finds the nearest 3-dimensional matches of a coordinate or coordinates in
    a set of catalog coordinates.

    This finds the 3-dimensional closest neighbor, which is only different
    from the on-sky distance if ``distance`` is set in either ``matchcoord``
    or ``catalogcoord``.

    Parameters
    ----------
    matchcoord : `~astropy.coordinates.BaseCoordinateFrame` or `~astropy.coordinates.SkyCoord`
        The coordinate(s) to match to the catalog.
    catalogcoord : `~astropy.coordinates.BaseCoordinateFrame` or `~astropy.coordinates.SkyCoord`
        The base catalog in which to search for matches. Typically this will
        be a coordinate object that is an array (i.e.,
        ``catalogcoord.isscalar == False``)
    nthneighbor : int, optional
        Which closest neighbor to search for.  Typically ``1`` is desired here,
        as that is correct for matching one set of coordinates to another.
        The next likely use case is ``2``, for matching a coordinate catalog
        against *itself* (``1`` is inappropriate because each point will find
        itself as the closest match).
    storekdtree : bool or str, optional
        If a string, will store the KD-Tree used for the computation
        in the ``catalogcoord``, as in ``catalogcoord.cache`` with the
        provided name.  This dramatically speeds up subsequent calls with the
        same catalog. If False, the KD-Tree is discarded after use.

    Returns
    -------
    CoordinateMatchResult
        A `~typing.NamedTuple` with attributes representing for each
        source in ``matchcoord`` the indices and angular and physical
        separations of the match in ``catalogcoord``.

    Notes
    -----
    This function requires `SciPy <https://www.scipy.org/>`_ to be installed
    or it will fail.
    """
    if catalogcoord.isscalar or len(catalogcoord) < 1:
        raise ValueError(
            "The catalog for coordinate matching cannot be a scalar or length-0."
        )

    kdt = _get_cartesian_kdtree(catalogcoord, storekdtree)

    # make sure coordinate systems match
    if isinstance(matchcoord, SkyCoord):
        matchcoord = matchcoord.transform_to(catalogcoord, merge_attributes=False)
    else:
        matchcoord = matchcoord.transform_to(catalogcoord)

    # make sure units match
    catunit = catalogcoord.cartesian.x.unit
    matchxyz = matchcoord.cartesian.xyz.to(catunit)

    matchflatxyz = matchxyz.reshape((3, np.prod(matchxyz.shape) // 3))
    # Querying NaN returns garbage
    if np.isnan(matchflatxyz.value).any():
        raise ValueError("Matching coordinates cannot contain NaN entries.")
    dist, idx = kdt.query(matchflatxyz.T, nthneighbor)

    if nthneighbor > 1:  # query gives 1D arrays if k=1, 2D arrays otherwise
        dist = dist[:, -1]
        idx = idx[:, -1]

    sep2d = catalogcoord[idx].separation(matchcoord)
    return CoordinateMatchResult(
        idx.reshape(matchxyz.shape[1:]),
        sep2d,
        dist.reshape(matchxyz.shape[1:]) * catunit,
    )


def match_coordinates_sky(
    matchcoord, catalogcoord, nthneighbor=1, storekdtree="kdtree_sky"
):
    """
    Finds the nearest on-sky matches of a coordinate or coordinates in
    a set of catalog coordinates.

    This finds the on-sky closest neighbor, which is only different from the
    3-dimensional match if ``distance`` is set in either ``matchcoord``
    or ``catalogcoord``.

    Parameters
    ----------
    matchcoord : `~astropy.coordinates.BaseCoordinateFrame` or `~astropy.coordinates.SkyCoord`
        The coordinate(s) to match to the catalog.
    catalogcoord : `~astropy.coordinates.BaseCoordinateFrame` or `~astropy.coordinates.SkyCoord`
        The base catalog in which to search for matches. Typically this will
        be a coordinate object that is an array (i.e.,
        ``catalogcoord.isscalar == False``)
    nthneighbor : int, optional
        Which closest neighbor to search for.  Typically ``1`` is desired here,
        as that is correct for matching one set of coordinates to another.
        The next likely use case is ``2``, for matching a coordinate catalog
        against *itself* (``1`` is inappropriate because each point will find
        itself as the closest match).
    storekdtree : bool or str, optional
        If a string, will store the KD-Tree used for the computation
        in the ``catalogcoord`` in ``catalogcoord.cache`` with the
        provided name.  This dramatically speeds up subsequent calls with the
        same catalog. If False, the KD-Tree is discarded after use.

    Returns
    -------
    CoordinateMatchResult
        A `~typing.NamedTuple` with attributes representing for each
        source in ``matchcoord`` the indices and angular and physical
        separations of the match in ``catalogcoord``. If either
        ``matchcoord`` or ``catalogcoord`` lack distances, the physical
        separation is the 3D distance on the unit sphere, rather than a
        true distance.

    Notes
    -----
    This function requires `SciPy <https://www.scipy.org/>`_ to be installed
    or it will fail.
    """
    if catalogcoord.isscalar or len(catalogcoord) < 1:
        raise ValueError(
            "The catalog for coordinate matching cannot be a scalar or length-0."
        )

    # send to catalog frame
    if isinstance(matchcoord, SkyCoord):
        newmatch = matchcoord.transform_to(catalogcoord, merge_attributes=False)
    else:
        newmatch = matchcoord.transform_to(catalogcoord)

    # strip out distance info
    match_urepr = newmatch.data.represent_as(UnitSphericalRepresentation)
    newmatch_u = newmatch.realize_frame(match_urepr)

    cat_urepr = catalogcoord.data.represent_as(UnitSphericalRepresentation)
    newcat_u = catalogcoord.realize_frame(cat_urepr)

    # Check for a stored KD-tree on the passed-in coordinate. Normally it will
    # have a distinct name from the "3D" one, so it's safe to use even though
    # it's based on UnitSphericalRepresentation.
    storekdtree = catalogcoord.cache.get(storekdtree, storekdtree)

    idx, sep2d, sep3d = match_coordinates_3d(
        newmatch_u, newcat_u, nthneighbor, storekdtree
    )
    # sep3d is *wrong* above, because the distance information was removed,
    # unless one of the catalogs doesn't have a real distance
    if not (
        isinstance(catalogcoord.data, UnitSphericalRepresentation)
        or isinstance(newmatch.data, UnitSphericalRepresentation)
    ):
        sep3d = catalogcoord[idx].separation_3d(newmatch)

    # update the kdtree on the actual passed-in coordinate
    if isinstance(storekdtree, str):
        catalogcoord.cache[storekdtree] = newcat_u.cache[storekdtree]
    elif storekdtree is True:
        # the old backwards-compatible name
        catalogcoord.cache["kdtree"] = newcat_u.cache["kdtree"]

    return CoordinateMatchResult(idx, sep2d, sep3d)


class CoordinateSearchResult(NamedTuple):
    """Results of searching close pairs between two sets of sources.

    See Also
    --------
    astropy.coordinates.search_around_3d
    astropy.coordinates.search_around_sky
    SkyCoord.search_around_3d
    SkyCoord.search_around_sky
    """

    indices_to_first_set: NDArray[np.int32] | NDArray[np.int64]
    """Indices of the elements of the found pairs in the first set of
    sources."""
    indices_to_second_set: NDArray[np.int32] | NDArray[np.int64]
    """Indices of the elements of the found pairs in the second set of
    sources."""
    angular_separation: Angle
    """The angular separations between the paired sources."""
    physical_separation: Quantity
    """The physical separations between the paired sources. If either of
    the source sets lack distances then the physical separations are
    computed assuming all coordinates are points on the unit sphere."""


def search_around_3d(coords1, coords2, distlimit, storekdtree="kdtree_3d"):
    """
    Searches for pairs of points that are at least as close as a specified
    distance in 3D space.

    This is intended for use on coordinate objects with arrays of coordinates,
    not scalars.  For scalar coordinates, it is better to use the
    ``separation_3d`` methods.

    Parameters
    ----------
    coords1 : `~astropy.coordinates.BaseCoordinateFrame` or `~astropy.coordinates.SkyCoord`
        The first set of coordinates, which will be searched for matches from
        ``coords2`` within ``seplimit``.  Must be a one-dimensional coordinate array.
    coords2 : `~astropy.coordinates.BaseCoordinateFrame` or `~astropy.coordinates.SkyCoord`
        The second set of coordinates, which will be searched for matches from
        ``coords1`` within ``seplimit``.  Must be a one-dimensional coordinate array.
    distlimit : `~astropy.units.Quantity` ['length']
        The physical radius to search within. It should be broadcastable to the
        same shape as ``coords1``.
    storekdtree : bool or str, optional
        If a string, will store the KD-Tree used in the search with the name
        ``storekdtree`` in ``coords2.cache``. This speeds up subsequent calls
        to this function. If False, the KD-Trees are not saved.

    Returns
    -------
    CoordinateSearchResult
        A `~typing.NamedTuple` with attributes representing the indices
        of the elements of found pairs in both source sets and angular
        and physical separations of the pairs.

    Notes
    -----
    This function requires `SciPy <https://www.scipy.org/>`_
    to be installed or it will fail.

    If you are using this function to search in a catalog for matches around
    specific points, the convention is for ``coords2`` to be the catalog, and
    ``coords1`` are the points to search around.  While these operations are
    mathematically the same if ``coords1`` and ``coords2`` are flipped, some of
    the optimizations may work better if this convention is obeyed.

    In the current implementation, the return values are always sorted in the
    same order as the ``coords1`` (so ``idx1`` is in ascending order).  This is
    considered an implementation detail, though, so it could change in a future
    release.
    """
    if coords1.ndim != 1 or coords2.ndim != 1:
        msg = "search_around_3d only supports 1-dimensional coordinate arrays."
        if coords1.isscalar or coords2.isscalar:
            msg += " With a scalar array, use ``coord1.separation(coord2) < seplimit``."
        raise ValueError(msg)

    kdt2 = _get_cartesian_kdtree(coords2, storekdtree)
    cunit = coords2.cartesian.x.unit

    # we convert coord1 to match coord2's frame.  We do it this way
    # so that if the conversion does happen, the KD tree of coord2 at least gets
    # saved. (by convention, coord2 is the "catalog" if that makes sense)
    coords1 = coords1.transform_to(coords2)

    kdt1 = _get_cartesian_kdtree(coords1, storekdtree, forceunit=cunit)
    idxs1 = []
    idxs2 = []

    if distlimit.isscalar:
        for i, matches in enumerate(
            kdt1.query_ball_tree(kdt2, distlimit.to_value(cunit))
        ):
            idxs1.extend(len(matches) * [i])
            idxs2.extend(matches)
    else:
        for i, (point, distance) in enumerate(zip(kdt1.data, distlimit, strict=True)):
            matches = kdt2.query_ball_point(point, distance.to_value(cunit))
            idxs1.extend(len(matches) * [i])
            idxs2.extend(matches)
    return CoordinateSearchResult(
        np.array(idxs1, dtype=int),
        np.array(idxs2, dtype=int),
        coords1[idxs1].separation(coords2[idxs2]),
        coords1[idxs1].separation_3d(coords2[idxs2]),
    )


def search_around_sky(coords1, coords2, seplimit, storekdtree="kdtree_sky"):
    """
    Searches for pairs of points that have an angular separation at least as
    close as a specified angle.

    This is intended for use on coordinate objects with arrays of coordinates,
    not scalars.  For scalar coordinates, it is better to use the ``separation``
    methods.

    Parameters
    ----------
    coords1 : coordinate-like
        The first set of coordinates, which will be searched for matches from
        ``coords2`` within ``seplimit``. Must be a one-dimensional coordinate array.
    coords2 : coordinate-like
        The second set of coordinates, which will be searched for matches from
        ``coords1`` within ``seplimit``. Must be a one-dimensional coordinate array.
    seplimit : `~astropy.units.Quantity` ['angle']
        The on-sky separation to search within. It should be broadcastable to the same
        shape as ``coords1``.
    storekdtree : bool or str, optional
        If a string, will store the KD-Tree used in the search with the name
        ``storekdtree`` in ``coords2.cache``. This speeds up subsequent calls
        to this function. If False, the KD-Trees are not saved.

    Returns
    -------
    CoordinateSearchResult
        A `~typing.NamedTuple` with attributes representing the indices
        of the elements of found pairs in both source sets and angular
        and physical separations of the pairs. If either set of sources
        lack distances, the physical separation is the 3D distance on
        the unit sphere, rather than a true distance.

    Notes
    -----
    This function requires `SciPy <https://www.scipy.org/>`_
    to be installed or it will fail.

    In the current implementation, the return values are always sorted in the
    same order as the ``coords1`` (so ``idx1`` is in ascending order).  This is
    considered an implementation detail, though, so it could change in a future
    release.
    """
    if coords1.ndim != 1 or coords2.ndim != 1:
        msg = "search_around_sky only supports 1-dimensional coordinate arrays."
        if coords1.isscalar or coords2.isscalar:
            msg += " With a scalar array, use ``coord1.separation(coord2) < seplimit``."
        raise ValueError(msg)

    # we convert coord1 to match coord2's frame.  We do it this way
    # so that if the conversion does happen, the KD tree of coord2 at least gets
    # saved. (by convention, coord2 is the "catalog" if that makes sense)
    coords1 = coords1.transform_to(coords2)

    # strip out distance info
    urepr1 = coords1.data.represent_as(UnitSphericalRepresentation)

    kdt1 = _get_cartesian_kdtree(coords1.realize_frame(urepr1), storekdtree)
    if storekdtree and coords2.cache.get(storekdtree):
        # just use the stored KD-Tree
        kdt2 = coords2.cache[storekdtree]
    else:
        # strip out distance info
        urepr2 = coords2.data.represent_as(UnitSphericalRepresentation)

        kdt2 = _get_cartesian_kdtree(coords2.realize_frame(urepr2), storekdtree)
        if storekdtree:
            coords2.cache["kdtree" if storekdtree is True else storekdtree] = kdt2

    idxs1 = []
    idxs2 = []

    if seplimit.isscalar:
        # this is the *cartesian* 3D distance that corresponds to the given angle
        r = (2 * np.sin(Angle(0.5 * seplimit))).value

        for i, matches in enumerate(kdt1.query_ball_tree(kdt2, r)):
            idxs1.extend(len(matches) * [i])
            idxs2.extend(matches)
    else:
        for i, (point, sep) in enumerate(zip(kdt1.data, seplimit, strict=True)):
            radius = (2 * np.sin(Angle(0.5 * sep))).value
            matches = kdt2.query_ball_point(point, radius)
            idxs1.extend(len(matches) * [i])
            idxs2.extend(matches)

    d2ds = coords1[idxs1].separation(coords2[idxs2])
    try:
        d3ds = coords1[idxs1].separation_3d(coords2[idxs2])
    except ValueError:
        # they don't have distances, so we just fall back on the cartesian
        # distance, computed from d2ds
        d3ds = 2 * np.sin(0.5 * d2ds)
    return CoordinateSearchResult(
        np.array(idxs1, dtype=int), np.array(idxs2, dtype=int), d2ds, d3ds
    )


def _get_cartesian_kdtree(coord, attrname_or_kdt="kdtree", forceunit=None):
    """
    This is a utility function to retrieve (and build/cache, if necessary)
    a 3D cartesian KD-Tree from various sorts of astropy coordinate objects.

    Parameters
    ----------
    coord : `~astropy.coordinates.BaseCoordinateFrame` or `~astropy.coordinates.SkyCoord`
        The coordinates to build the KD-Tree for.
    attrname_or_kdt : bool or str or KDTree
        If a string, will store the KD-Tree used for the computation in the
        ``coord``, in ``coord.cache`` with the provided name. If given as a
        KD-Tree, it will just be used directly.
    forceunit : unit or None
        If a unit, the cartesian coordinates will convert to that unit before
        being put in the KD-Tree.  If None, whatever unit it's already in
        will be used

    Returns
    -------
    kdt : `~scipy.spatial.KDTree`
        The KD-Tree representing the 3D cartesian representation of the input
        coordinates.
    """
    from scipy.spatial import KDTree

    if attrname_or_kdt is True:  # backwards compatibility for pre v0.4
        attrname_or_kdt = "kdtree"

    # figure out where any cached KDTree might be
    if isinstance(attrname_or_kdt, str):
        kdt = coord.cache.get(attrname_or_kdt, None)
        if kdt is not None and not isinstance(kdt, KDTree):
            raise TypeError(
                f'The `attrname_or_kdt` "{attrname_or_kdt}" is not a scipy KD tree!'
            )
    elif isinstance(attrname_or_kdt, KDTree):
        kdt = attrname_or_kdt
        attrname_or_kdt = None
    elif not attrname_or_kdt:
        kdt = None
    else:
        raise TypeError(
            "Invalid `attrname_or_kdt` argument for KD-Tree:" + str(attrname_or_kdt)
        )

    if kdt is None:
        # need to build the cartesian KD-tree for the catalog
        if forceunit is None:
            cartxyz = coord.cartesian.xyz
        else:
            cartxyz = coord.cartesian.xyz.to(forceunit)
        flatxyz = cartxyz.reshape((3, np.prod(cartxyz.shape) // 3))
        # There should be no NaNs in the kdtree data.
        if np.isnan(flatxyz.value).any():
            raise ValueError("Catalog coordinates cannot contain NaN entries.")
        # Not obvious if compact_nodes=False, balanced_tree=False is still needed but
        # we stay backwards-compatible with previous versions of `astropy` for now.
        kdt = KDTree(flatxyz.value.T, compact_nodes=False, balanced_tree=False)

    if attrname_or_kdt:
        # cache the kdtree in `coord`
        coord.cache[attrname_or_kdt] = kdt

    return kdt



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings
from pathlib import Path

from astropy import log
from astropy.io.fits import getdata
from astropy.utils.decorators import deprecated_renamed_argument
from astropy.visualization.mpl_normalize import simple_norm

__all__ = ["fits2bitmap", "main"]


@deprecated_renamed_argument(["min_cut", "max_cut"], ["vmin", "vmax"], ["6.1", "6.1"])
def fits2bitmap(
    filename,
    ext=0,
    out_fn=None,
    stretch="linear",
    power=1.0,
    asinh_a=0.1,
    vmin=None,
    vmax=None,
    min_percent=None,
    max_percent=None,
    percent=None,
    cmap="Greys_r",
):
    """
    Create a bitmap file from a FITS image, applying a stretching
    transform between minimum and maximum cut levels and a matplotlib
    colormap.

    Parameters
    ----------
    filename : str | PathLike
        The filename of the FITS file.
    ext : int
        FITS extension name or number of the image to convert. The
        default is 0.
    out_fn : str | PathLike
        The filename of the output bitmap image. The type of bitmap is
        determined by the filename extension (e.g. '.jpg', '.png'). The
        default is a PNG file with the same name as the FITS file.
    stretch : {'linear', 'sqrt', 'power', log', 'asinh'}
        The stretching function to apply to the image. The default is
        'linear'.
    power : float, optional
        The power index for ``stretch='power'``. The default is 1.0.
    asinh_a : float, optional
        For ``stretch='asinh'``, the value where the asinh curve
        transitions from linear to logarithmic behavior, expressed as a
        fraction of the normalized image. Must be in the range between 0
        and 1. The default is 0.1.
    vmin : float, optional
        The pixel value of the minimum cut level. Data values less
        than ``vmin`` will set to ``vmin`` before stretching the
        image. The default is the image minimum. ``vmin`` overrides
        ``min_percent``.
    vmax : float, optional
        The pixel value of the maximum cut level. Data values greater
        than ``vmax`` will set to ``vmax`` before stretching the
        image. The default is the image maximum. ``vmax`` overrides
        ``max_percent``.
    min_percent : float, optional
        The percentile value used to determine the pixel value of
        minimum cut level. The default is 0.0. ``min_percent`` overrides
        ``percent``.
    max_percent : float, optional
        The percentile value used to determine the pixel value of
        maximum cut level. The default is 100.0. ``max_percent``
        overrides ``percent``.
    percent : float, optional
        The percentage of the image values used to determine the pixel
        values of the minimum and maximum cut levels. The lower cut
        level will set at the ``(100 - percent) / 2`` percentile, while
        the upper cut level will be set at the ``(100 + percent) / 2``
        percentile. The default is 100.0. ``percent`` is ignored if
        either ``min_percent`` or ``max_percent`` is input.
    cmap : str
        The matplotlib color map name. The default is 'Greys_r'.
    """
    import matplotlib as mpl
    import matplotlib.image as mimg

    # __main__ gives ext as a string
    try:
        ext = int(ext)
    except ValueError:
        pass

    try:
        image = getdata(filename, ext)
    except Exception as e:
        log.critical(e)
        return 1

    if image.ndim != 2:
        log.critical(f"data in FITS extension {ext} is not a 2D array")

    filename = Path(filename)

    if out_fn is None:
        out_fn = filename.with_suffix("")
        # If the filename ends with .fits.*, remove the * extension.
        if out_fn.suffix == ".fits":
            out_fn = out_fn.with_suffix("")
        out_fn = out_fn.with_suffix(out_fn.suffix + ".png")
    else:
        out_fn = Path(out_fn)
    out_format = out_fn.suffix[1:]

    if cmap not in mpl.colormaps:
        log.critical(f"{cmap} is not a valid matplotlib colormap name.")
        return 1

    norm = simple_norm(
        image,
        stretch=stretch,
        power=power,
        asinh_a=asinh_a,
        vmin=vmin,
        vmax=vmax,
        min_percent=min_percent,
        max_percent=max_percent,
        percent=percent,
    )

    mimg.imsave(out_fn, norm(image), cmap=cmap, origin="lower", format=out_format)
    log.info(f"Saved file to {out_fn}.")


def main(args=None):
    import argparse

    parser = argparse.ArgumentParser(
        description="Create a bitmap file from a FITS image."
    )
    # TODO: pass suggest_on_error as kwarg when PYTHON_LT_14 is dropped
    parser.suggest_on_error = True

    # the mutually exclusive groups can be removed when the deprecated
    # min_cut and max_cut are removed
    vmin_group = parser.add_mutually_exclusive_group()
    vmax_group = parser.add_mutually_exclusive_group()
    parser.add_argument(
        "-e",
        "--ext",
        metavar="hdu",
        default=0,
        help="Specify the HDU extension number or name (Default is 0).",
    )
    parser.add_argument(
        "-o",
        metavar="filename",
        type=str,
        default=None,
        help=(
            "Filename for the output image (Default is a "
            "PNG file with the same name as the FITS file)."
        ),
    )
    parser.add_argument(
        "--stretch",
        type=str,
        default="linear",
        help=(
            'Type of image stretching ("linear", "sqrt", '
            '"power", "log", or "asinh") (Default is "linear").'
        ),
    )
    parser.add_argument(
        "--power",
        type=float,
        default=1.0,
        help='Power index for "power" stretching (Default is 1.0).',
    )
    parser.add_argument(
        "--asinh_a",
        type=float,
        default=0.1,
        help=(
            "The value in normalized image where the asinh "
            "curve transitions from linear to logarithmic "
            'behavior (used only for "asinh" stretch) '
            "(Default is 0.1)."
        ),
    )
    vmin_group.add_argument(
        "--vmin",
        type=float,
        default=None,
        help="The pixel value of the minimum cut level (Default is the image minimum).",
    )
    vmax_group.add_argument(
        "--vmax",
        type=float,
        default=None,
        help="The pixel value of the maximum cut level (Default is the image maximum).",
    )
    vmin_group.add_argument(
        "--min_cut",
        type=float,
        default=None,
        help="The pixel value of the minimum cut level (Deprecated, use vmin instead; default is the image minimum).",
    )
    vmax_group.add_argument(
        "--max_cut",
        type=float,
        default=None,
        help="The pixel value of the maximum cut level (Deprecated, use vmax instead; default is the image maximum).",
    )
    parser.add_argument(
        "--min_percent",
        type=float,
        default=None,
        help=(
            "The percentile value used to determine the "
            "minimum cut level (Default is 0)."
        ),
    )
    parser.add_argument(
        "--max_percent",
        type=float,
        default=None,
        help=(
            "The percentile value used to determine the "
            "maximum cut level (Default is 100)."
        ),
    )
    parser.add_argument(
        "--percent",
        type=float,
        default=None,
        help=(
            "The percentage of the image values used to "
            "determine the pixel values of the minimum and "
            "maximum cut levels (Default is 100)."
        ),
    )
    parser.add_argument(
        "--cmap",
        metavar="colormap_name",
        type=str,
        default="Greys_r",
        help='matplotlib color map name (Default is "Greys_r").',
    )
    parser.add_argument(
        "filename", nargs="+", help="Path to one or more FITS files to convert"
    )
    args = parser.parse_args(args)

    if args.min_cut is not None:
        warnings.warn('The "--min_cut" argument is deprecated. Use "--vmin" instead.')
        args.vmin = args.min_cut

    if args.max_cut is not None:
        warnings.warn('The "--max_cut" argument is deprecated. Use "--vmax" instead.')
        args.vmax = args.max_cut

    for filename in args.filename:
        fits2bitmap(
            filename,
            ext=args.ext,
            out_fn=args.o,
            stretch=args.stretch,
            vmin=args.vmin,
            vmax=args.vmax,
            min_percent=args.min_percent,
            max_percent=args.max_percent,
            percent=args.percent,
            power=args.power,
            asinh_a=args.asinh_a,
            cmap=args.cmap,
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Hubble parameter trait.

This is private API. See `~astropy.cosmology.traits` for public API.
"""

__all__ = ["HubbleParameter"]

from collections.abc import Callable
from functools import cached_property
from typing import Any

import numpy as np
from numpy.typing import ArrayLike, NDArray

import astropy.units as u
from astropy import constants as const
from astropy.cosmology._src.typing import FArray
from astropy.cosmology._src.utils import deprecated_keywords
from astropy.units import Quantity


class HubbleParameter:
    """The object has attributes and methods for the Hubble parameter."""

    H0: Quantity
    """Hubble Parameter at redshift 0."""

    efunc: Callable[[Any], NDArray[Any]]

    inv_efunc: Callable[[Any], FArray | float]

    @deprecated_keywords("z", since="7.0")
    def H(self, z: Quantity | ArrayLike) -> Quantity:
        """Hubble parameter at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

        Returns
        -------
        H : Quantity ['frequency']
            Hubble parameter at each input redshift.
        """
        return self.H0 * self.efunc(z)

    @cached_property
    def h(self) -> np.floating:
        """Dimensionless Hubble constant: h = H_0 / 100 [km/sec/Mpc]."""
        return self.H0.to_value("km/(s Mpc)") / 100.0

    @cached_property
    def hubble_time(self) -> u.Quantity:
        """Hubble time."""
        return (1 / self.H0).to(u.Gyr)

    @cached_property
    def hubble_distance(self) -> u.Quantity:
        """Hubble distance."""
        return (const.c / self.H0).to(u.Mpc)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Classes for handling metadata."""

__all__ = ["MetaAttribute", "MetaData"]

import inspect
from collections import OrderedDict
from collections.abc import Mapping
from copy import deepcopy
from dataclasses import is_dataclass


class MetaData:
    """
    A descriptor for classes that have a ``meta`` property.

    This can be set to any valid :class:`~collections.abc.Mapping`.

    Parameters
    ----------
    doc : `str`, optional
        Documentation for the attribute of the class.
        Default is ``""``.

        .. versionadded:: 1.2

    copy : `bool`, optional
        If ``True`` the value is deepcopied before setting, otherwise it
        is saved as reference.
        Default is ``True``.

        .. versionadded:: 1.2

    default_factory : Callable[[], Mapping], optional keyword-only
        The factory to use to create the default value of the ``meta``
        attribute.  This must be a callable that returns a `Mapping` object.
        Default is `OrderedDict`, creating an empty `OrderedDict`.

        .. versionadded:: 6.0

    Examples
    --------
    ``MetaData`` can be used as a descriptor to define a ``meta`` attribute`.

        >>> class Foo:
        ...     meta = MetaData()
        ...     def __init__(self, meta=None):
        ...         self.meta = meta

    ``Foo`` can be instantiated with a ``meta`` argument.

        >>> foo = Foo(meta={'a': 1, 'b': 2})
        >>> foo.meta
        {'a': 1, 'b': 2}

    The default value of ``meta`` is an empty :class:`~collections.OrderedDict`.
    This can be set by passing ``None`` to the ``meta`` argument.

        >>> foo = Foo()
        >>> foo.meta
        OrderedDict()

    If an :class:`~collections.OrderedDict` is not a good default metadata type then
    the ``default_factory`` keyword can be used to set the default to a different
    `Mapping` type, when the class is defined.'

        >>> class Bar:
        ...     meta = MetaData(default_factory=dict)
        ...     def __init__(self, meta=None):
        ...         self.meta = meta

        >>> Bar().meta
        {}

    When accessed from the class ``.meta`` returns `None` since metadata is
    on the class' instances, not the class itself.

        >>> print(Foo.meta)
        None
    """

    def __init__(self, doc="", copy=True, *, default_factory=OrderedDict):
        self.__doc__ = doc
        self.copy = copy
        self._default_factory = default_factory

    @property
    def default_factory(self):
        return self._default_factory

    def __get__(self, instance, owner):
        # class attribute access. Often, descriptors just return `self`, but if the
        # owning class is a `dataclass`, the expectation is that the default is
        # returned. In our case, this is None, triggering the creation of a dict-like in
        # `__set__`.
        if instance is None:
            return None
        # instance attribute access
        if not hasattr(instance, "_meta"):
            self.__set__(instance, None)
        return instance._meta

    def __set__(self, instance, value):
        # The 'default' value is `None`, but we want to set it to an empty `Mapping`
        # if it is `None` so that we can always assume it is a `Mapping` and not have
        # to check for `None` everywhere.
        if value is None:
            value = self.default_factory()
        # We don't want to allow setting the meta attribute to a non-dict-like object.
        # NOTE: with mypyc compilation this can be removed.
        elif not isinstance(value, Mapping):
            raise TypeError("meta attribute must be dict-like")
        # This is called when the dataclass is instantiated with a `meta` argument.
        else:
            value = deepcopy(value) if self.copy else value

        if is_dataclass(instance) and instance.__dataclass_params__.frozen:
            object.__setattr__(instance, "_meta", value)
        else:
            instance._meta = value


class MetaAttribute:
    """
    Descriptor to define custom attribute which gets stored in the object
    ``meta`` dict and can have a defined default.

    This descriptor is intended to provide a convenient way to add attributes
    to a subclass of a complex class such as ``Table`` or ``NDData``.

    This requires that the object has an attribute ``meta`` which is a
    dict-like object.  The value of the MetaAttribute will be stored in a
    new dict meta['__attributes__'] that is created when required.

    Classes that define MetaAttributes are encouraged to support initializing
    the attributes via the class ``__init__``.  For example::

        for attr in list(kwargs):
            descr = getattr(self.__class__, attr, None)
            if isinstance(descr, MetaAttribute):
                setattr(self, attr, kwargs.pop(attr))

    The name of a ``MetaAttribute`` cannot be the same as any of the following:

    - Keyword argument in the owner class ``__init__``
    - Method or attribute of the "parent class", where the parent class is
      taken to be ``owner.__mro__[1]``.

    Parameters
    ----------
    default : Any, optional
        Default value for the attribute, by default `None`.
    """

    def __init__(self, default=None):
        self.default = default

    def __get__(self, instance, owner):
        # When called without an instance, return self to allow access
        # to descriptor attributes.
        if instance is None:
            return self

        # If default is None and value has not been set already then return None
        # without doing touching meta['__attributes__'] at all. This helps e.g.
        # with the Table._hidden_columns attribute so it doesn't auto-create
        # meta['__attributes__'] always.
        if self.default is None and self.name not in instance.meta.get(
            "__attributes__", {}
        ):
            return None

        # Get the __attributes__ dict and create if not there already.
        attributes = instance.meta.setdefault("__attributes__", {})
        try:
            value = attributes[self.name]
        except KeyError:
            if self.default is not None:
                attributes[self.name] = deepcopy(self.default)
            # Return either specified default or None
            value = attributes.get(self.name)
        return value

    def __set__(self, instance, value):
        # Get the __attributes__ dict and create if not there already.
        attributes = instance.meta.setdefault("__attributes__", {})
        attributes[self.name] = value

    def __delete__(self, instance):
        # Remove this attribute from meta['__attributes__'] if it exists.
        if "__attributes__" in instance.meta:
            attrs = instance.meta["__attributes__"]
            if self.name in attrs:
                del attrs[self.name]
            # If this was the last attribute then remove the meta key as well
            if not attrs:
                del instance.meta["__attributes__"]

    def __set_name__(self, owner, name):
        params = [
            param.name
            for param in inspect.signature(owner).parameters.values()
            if param.kind
            not in (inspect.Parameter.VAR_KEYWORD, inspect.Parameter.VAR_POSITIONAL)
        ]

        # Reject names from existing params or best guess at parent class
        if name in params or hasattr(owner.__mro__[1], name):
            raise ValueError(f"{name} not allowed as {self.__class__.__name__}")

        self.name = name

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name} default={self.default}>"



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This subpackage contains classes and functions for work with time series.
"""

from astropy.timeseries import io
from astropy.timeseries.binned import *
from astropy.timeseries.core import *
from astropy.timeseries.downsample import *
from astropy.timeseries.periodograms import *
from astropy.timeseries.sampled import *



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest
from hypothesis import given
from hypothesis.extra.numpy import basic_indices
from numpy.testing import assert_equal

from astropy.utils.exceptions import AstropyDeprecationWarning
from astropy.utils.shapes import check_broadcast, simplify_basic_index, unbroadcast


def test_check_broadcast_deprecation():
    with pytest.warns(AstropyDeprecationWarning):
        check_broadcast((1,), (2,))


@pytest.mark.filterwarnings("ignore")
def test_check_broadcast():
    assert check_broadcast((10, 1), (3,)) == (10, 3)
    assert check_broadcast((10, 1), (3,), (4, 1, 1, 3)) == (4, 1, 10, 3)
    with pytest.raises(ValueError):
        check_broadcast((10, 2), (3,))

    with pytest.raises(ValueError):
        check_broadcast((10, 1), (3,), (4, 1, 2, 3))


def test_unbroadcast():
    x = np.array([1, 2, 3])
    y = np.broadcast_to(x, (2, 4, 3))
    z = unbroadcast(y)
    assert z.shape == (3,)
    np.testing.assert_equal(z, x)

    x = np.ones((3, 5))
    y = np.broadcast_to(x, (5, 3, 5))
    z = unbroadcast(y)
    assert z.shape == (3, 5)


TEST_SHAPE = (13, 16, 4, 90)


class TestSimplifyBasicIndex:
    # We use a class here so that we can allocate the data once and for all to
    # speed up the testing.

    def setup_class(self):
        self.shape = TEST_SHAPE
        self.data = np.random.random(TEST_SHAPE)

    @given(basic_indices(TEST_SHAPE))
    def test_indexing(self, index):
        new_index = simplify_basic_index(index, shape=self.shape)
        assert_equal(self.data[index], self.data[new_index])
        assert isinstance(new_index, tuple)
        assert len(new_index) == len(self.shape)
        for idim, idx in enumerate(new_index):
            assert isinstance(idx, (slice, int))
            if isinstance(idx, int):
                assert idx >= 0
            else:
                assert isinstance(idx.start, int)
                assert idx.start >= 0
                assert idx.start < TEST_SHAPE[idim]
                if idx.stop is not None:
                    assert isinstance(idx.stop, int)
                    assert idx.stop >= 0
                    assert idx.stop <= TEST_SHAPE[idim]
                assert isinstance(idx.step, int)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest
from numpy.testing import assert_allclose

from astropy import units as u
from astropy.io import ascii

from .common import (
    setup_function,  # noqa: F401
    teardown_function,  # noqa: F401
)


def read_table1(readme, data):
    reader = ascii.Cds(readme)
    return reader.read(data)


def read_table2(readme, data):
    reader = ascii.get_reader(reader_cls=ascii.Cds, readme=readme)
    reader.outputter = ascii.TableOutputter()
    return reader.read(data)


def read_table3(readme, data):
    return ascii.read(data, readme=readme)


def test_description():
    readme = "data/cds/description/ReadMe"
    data = "data/cds/description/table.dat"
    for read_table in (read_table1, read_table2, read_table3):
        table = read_table(readme, data)
        assert len(table) == 2
        assert table["Cluster"].description == "Cluster name"
        assert table["Star"].description == ""
        assert table["Wave"].description == "wave ? Wavelength in Angstroms"
        assert table["El"].description == "a"
        assert table["ion"].description == "- Ionization stage (1 for neutral element)"
        assert table["loggf"].description == (
            "log10 of the gf value - logarithm base 10 of stat. weight times "
            "oscillator strength"
        )
        assert table["EW"].description == "Equivalent width (in mA)"
        assert (
            table["Q"].description
            == "DAOSPEC quality parameter Q (large values are bad)"
        )


def test_multi_header():
    readme = "data/cds/multi/ReadMe"
    data = "data/cds/multi/lhs2065.dat"
    for read_table in (read_table1, read_table2, read_table3):
        table = read_table(readme, data)
        assert len(table) == 18
        assert_allclose(table["Lambda"][-1], 6479.32)
        assert table["Fnu"][-1] == "0.285937"
    data = "data/cds/multi/lp944-20.dat"
    for read_table in (read_table1, read_table2, read_table3):
        table = read_table(readme, data)
        assert len(table) == 18
        assert_allclose(table["Lambda"][0], 6476.09)
        assert table["Fnu"][-1] == "0.489005"


def test_glob_header():
    readme = "data/cds/glob/ReadMe"
    data = "data/cds/glob/lmxbrefs.dat"
    for read_table in (read_table1, read_table2, read_table3):
        table = read_table(readme, data)
        assert len(table) == 291
        assert table["Name"][-1] == "J1914+0953"
        assert table["BibCode"][-2] == "2005A&A...432..235R"


def test_header_from_readme():
    """Test reading VizieR data with accompanying ReadMe file.

    We test several things here to make the best use of the included files:
    - "table1.dat" is a standard table.
    - "table5.dat.gz" is a gzipped table. When downloaded from VizieR, large tables come
      gzipped by default, but the name in the ReadMe does not include the ".gz" suffix.
      So, we check here that the zipped table is found. Regression test for #6549.
    """
    r = ascii.Cds("data/vizier/ReadMe")
    table = r.read("data/vizier/table1.dat")
    assert len(r.data.data_lines) == 15
    assert len(table) == 15
    assert len(table.keys()) == 18
    Bmag = [
        14.79,
        15.00,
        14.80,
        12.38,
        12.36,
        12.24,
        13.75,
        13.65,
        13.41,
        11.59,
        11.68,
        11.53,
        13.92,
        14.03,
        14.18,
    ]
    for i, val in enumerate(table.field("Bmag")):
        assert val == Bmag[i]

    table = r.read("data/vizier/table5.dat.gz")
    assert len(r.data.data_lines) == 49
    assert len(table) == 49
    assert len(table.keys()) == 10
    Q = [
        0.289,
        0.325,
        0.510,
        0.577,
        0.539,
        0.390,
        0.957,
        0.736,
        1.435,
        1.117,
        1.473,
        0.808,
        1.416,
        2.209,
        0.617,
        1.046,
        1.604,
        1.419,
        1.431,
        1.183,
        1.210,
        1.005,
        0.706,
        0.665,
        0.340,
        0.323,
        0.391,
        0.280,
        0.343,
        0.369,
        0.495,
        0.828,
        1.113,
        0.499,
        1.038,
        0.260,
        0.863,
        1.638,
        0.479,
        0.232,
        0.627,
        0.671,
        0.371,
        0.851,
        0.607,
        -9.999,
        1.958,
        1.416,
        0.949,
    ]
    for i, val in enumerate(table.field("Q")):
        if val is np.ma.masked:
            # text value for a missing value in that table
            assert Q[i] == -9.999
        else:
            assert val == Q[i]


@pytest.mark.parametrize("reader_cls", (ascii.Cds, ascii.Mrt))
def test_cds_units(reader_cls):
    from astropy import units

    data_and_readme = "data/cds.dat"
    reader = ascii.get_reader(reader_cls)
    table = reader.read(data_and_readme)
    # column unit is GMsun (giga solar masses)
    # make sure this is parsed correctly, not as a "string" unit
    assert table["Fit"].to(units.solMass).unit == units.solMass


@pytest.mark.parametrize("reader_cls", (ascii.Cds, ascii.Mrt))
def test_cds_function_units(reader_cls):
    data_and_readme = "data/cdsFunctional.dat"
    reader = ascii.get_reader(reader_cls)
    table = reader.read(data_and_readme)
    assert table["logg"].unit == u.dex(u.cm / u.s**2)
    assert table["logTe"].unit == u.dex(u.K)
    assert table["Mass"].unit == u.Msun
    assert table["e_Mass"].unit == u.Msun
    assert table["Age"].unit == u.Myr
    assert table["e_Age"].unit == u.Myr


@pytest.mark.parametrize("reader_cls", (ascii.Cds, ascii.Mrt))
def test_cds_function_units2(reader_cls):
    # This one includes some dimensionless dex.
    data_and_readme = "data/cdsFunctional2.dat"
    reader = ascii.get_reader(reader_cls)
    table = reader.read(data_and_readme)
    assert table["Teff"].unit == u.K
    assert table["logg"].unit == u.dex(u.cm / u.s**2)
    assert table["vturb"].unit == u.km / u.s
    assert table["[Fe/H]"].unit == u.dex(u.one)
    assert table["e_[Fe/H]"].unit == u.dex(u.one)
    assert_allclose(
        table["[Fe/H]"].to(u.one), 10.0 ** (np.array([-2.07, -1.50, -2.11, -1.64]))
    )


def test_cds_ignore_nullable():
    # Make sure CDS reader_cls does not ignore nullabilty for columns
    # with a limit specifier
    readme = "data/cds/null/ReadMe"
    data = "data/cds/null/table.dat"
    r = ascii.Cds(readme)
    r.read(data)
    assert r.header.cols[6].description == "Temperature class codified (10)"
    assert r.header.cols[8].description == "Luminosity class codified (11)"
    assert r.header.cols[5].description == "Pericenter position angle (18)"


def test_cds_no_whitespace():
    # Make sure CDS reader_cls only checks null values when an '=' symbol is present,
    # and read description text even if there is no whitespace after '?'.
    readme = "data/cds/null/ReadMe1"
    data = "data/cds/null/table1.dat"
    r = ascii.Cds(readme)
    r.read(data)
    assert r.header.cols[6].description == "Temperature class codified (10)"
    assert r.header.cols[6].null == ""
    assert r.header.cols[7].description == "Equivalent width (in mA)"
    assert r.header.cols[7].null == "-9.9"
    assert r.header.cols[10].description == (
        "DAOSPEC quality parameter Q (large values are bad)"
    )
    assert r.header.cols[10].null == "-9.999"


def test_cds_order():
    # Make sure CDS reader_cls does not ignore order specifier that maybe present after
    # the null specifier '?'
    readme = "data/cds/null/ReadMe1"
    data = "data/cds/null/table1.dat"
    r = ascii.Cds(readme)
    r.read(data)
    assert r.header.cols[5].description == "Catalogue Identification Number"
    assert r.header.cols[8].description == "Another equivalent width (in mA)"
    assert r.header.cols[9].description == "Luminosity class codified (11)"
